class NonNumericOperandError(TypeError):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class EmptyExpressionError(ValueError):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class ExpressionSyntaxError(ValueError):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class MalformedExpressionError(ValueError):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class RuleTreeInvalidError(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)



import re
from functools import wraps

from yamale.util import isstr
from yamale.validators import Boolean, DefaultValidators, Enum, Integer, Number, Validator


def all_validators() -> dict:
    validators = DefaultValidators.copy()
    validators[StringBool.tag] = StringBool
    validators[StringInt.tag] = StringInt
    validators[StringFloat.tag] = StringFloat
    validators[ReplicaList.tag] = ReplicaList
    validators[StringEnum.tag] = StringEnum

    return validators


def catch_value_errors(validate):
    """method decorator to catch ValueErrors for casts and return an error"""

    @wraps(validate)
    def wrapper(self, value):
        try:
            return validate(self, value)
        except ValueError:
            return [f"'{value}' could not be casted to {self.tag[2:]}"]

    return wrapper


class StringBool(Boolean):
    tag = "s_bool"

    @catch_value_errors
    def validate(self, value):
        value = str(value)
        if value.lower() not in ["false", "true"]:
            raise ValueError
        return super().validate(bool(value))


class StringInt(Integer):
    tag = "s_int"

    @catch_value_errors
    def validate(self, value):
        return super().validate(int(value))


class StringFloat(Number):
    tag = "s_float"

    @catch_value_errors
    def validate(self, value):
        return super().validate(float(value))


class StringEnum(Enum):
    tag = "s_enum"

    def __init__(self, *args, case_sensitive: bool = True, **kwargs):
        if not case_sensitive:
            args = [arg.lower() for arg in args]
        super().__init__(*args, **kwargs)

    @catch_value_errors
    def validate(self, value):
        if not isinstance(value, str):
            raise TypeError(f"Value {value} has to be a string, but is {type(value).__name__}")
        return super().validate(value.lower())


class ReplicaList(Validator):
    """
    Validates a list of replicas in the form of '<broker_id>:<partition>' (e.g. `'0:0,1:1,2:2'`).
    Empty string for empty list or '*' for all replicas area also valid values.
    """

    tag = "replica_list"

    def _is_valid(self, value) -> bool:
        if not isstr(value):
            return False
        if value == "" or value == "*":
            return True
        for pair in value.split(","):
            if not re.fullmatch(r"\d+:\d+", pair):
                return False
        return True

    def fail(self, value):
        return f"could not build dict from this: {value}"



import json
from collections import namedtuple
from functools import lru_cache
from typing import Dict

import fastavro
import requests

SchemaPair = namedtuple("SchemaPair", ["original_schema", "parsed_schema"])


class SchemaRegistryClient:
    def __init__(self, schema_registry_uri: str):
        self.schema_registry_uri = schema_registry_uri

    @lru_cache(maxsize=100)
    def get_schema_from_id(self, schema_id: int) -> SchemaPair:
        url = f"{self.schema_registry_uri}/schemas/ids/{schema_id}"
        response = requests.get(url)
        response.raise_for_status()
        schema: Dict = json.loads(response.json()["schema"])
        return SchemaPair(schema, fastavro.schema.parse_schema(schema))



import logging
import re
from enum import Enum
from itertools import islice
from logging import Logger
from typing import TYPE_CHECKING, Dict, List, Union

import confluent_kafka
import pendulum
from click import BadParameter
from confluent_kafka.admin import ConfigResource
from confluent_kafka.admin import TopicMetadata as ConfluentTopic
from confluent_kafka.cimpl import NewTopic, TopicPartition
from pykafka.topic import Topic as PyKafkaTopic

from esque.clients.consumer import MessageConsumer
from esque.config import PING_GROUP_ID, Config
from esque.errors import MessageEmptyException, raise_for_kafka_error
from esque.helpers import ensure_kafka_future_done, invalidate_cache_after
from esque.resources.topic import Partition, PartitionInfo, Topic, TopicDiff

logger: Logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from esque.cluster import Cluster


class ClientTypes(Enum):
    Confluent = "Confluent"
    PyKafka = "PyKafka"


ClientType = Union[ConfluentTopic, PyKafkaTopic]


class TopicController:
    def __init__(self, cluster: "Cluster", config: Config):
        self.cluster: "Cluster" = cluster
        self.config = config

    def _get_client_topic(self, topic_name: str, client_type: ClientTypes) -> ClientType:
        confluent_topics = self.cluster.confluent_client.list_topics(topic=topic_name, timeout=10).topics
        # Confluent returns a list of requested topics with an Error as result if the topic doesn't exist
        topic_metadata: ConfluentTopic = confluent_topics[topic_name]
        raise_for_kafka_error(topic_metadata.error)
        if client_type == ClientTypes.Confluent:
            return confluent_topics[topic_name]
        elif client_type == ClientTypes.PyKafka:
            # at least PyKafka does it's own caching, so we don't have to bother
            pykafka_topics = self.cluster.pykafka_client.cluster.topics
            return pykafka_topics[topic_name]
        else:
            raise BadParameter(f"TopicType needs to be part of {ClientTypes}", param=client_type)

    def list_topics(
        self,
        *,
        search_string: str = None,
        sort: bool = True,
        hide_internal: bool = True,
        get_topic_objects: bool = True,
    ) -> List[Topic]:
        self.cluster.confluent_client.poll(timeout=1)
        topic_results = self.cluster.confluent_client.list_topics().topics.values()
        topic_names = [t.topic for t in topic_results]
        if search_string:
            topic_names = [topic for topic in topic_names if re.match(search_string, topic)]
        if hide_internal:
            topic_names = [topic for topic in topic_names if not topic.startswith("__")]
        if sort:
            topic_names = sorted(topic_names)

        if get_topic_objects:
            topics = list(map(self.get_cluster_topic, topic_names))
        else:
            topics = list(map(self.get_local_topic, topic_names))
        return topics

    @invalidate_cache_after
    def create_topics(self, topics: List[Topic]):
        for topic in topics:
            partitions = (
                topic.num_partitions if topic.num_partitions is not None else self.config.default_num_partitions
            )
            replicas = (
                topic.replication_factor
                if topic.replication_factor is not None
                else self.config.default_replication_factor
            )
            new_topic = NewTopic(
                topic.name, num_partitions=partitions, replication_factor=replicas, config=topic.config
            )
            future_list = self.cluster.confluent_client.create_topics([new_topic], operation_timeout=60)
            ensure_kafka_future_done(next(islice(future_list.values(), 1)))

    @invalidate_cache_after
    def alter_configs(self, topics: List[Topic]):
        for topic in topics:
            altered_config = self._get_altered_config(topic)
            config_resource = ConfigResource(ConfigResource.Type.TOPIC, topic.name, altered_config)
            future_list = self.cluster.confluent_client.alter_configs([config_resource])
            ensure_kafka_future_done(next(islice(future_list.values(), 1)))

    def _get_altered_config(self, topic: Topic) -> Dict[str, str]:
        cluster_topic = self.get_cluster_topic(topic.name)
        current_config = cluster_topic.config.items()
        altered_config = {}
        for name, value in current_config:
            if name in topic.config:
                altered_config[name] = topic.config[name]
                continue
            altered_config[name] = value
        return altered_config

    @invalidate_cache_after
    def delete_topic(self, topic: Topic):
        future = self.cluster.confluent_client.delete_topics([topic.name])[topic.name]
        ensure_kafka_future_done(future)

    def get_cluster_topic(self, topic_name: str) -> Topic:
        """Convenience function getting an existing topic based on topic_name"""
        return self.update_from_cluster(Topic(topic_name))

    def get_local_topic(self, topic_name: str) -> Topic:
        return Topic(topic_name)

    def get_offsets_closest_to_timestamp(
        self, group_id: str, topic_name: str, timestamp_limit: pendulum
    ) -> Dict[int, int]:
        topic = self.get_cluster_topic(topic_name=topic_name)
        config = Config.get_instance().create_confluent_config()
        config.update({"group.id": group_id})
        consumer = confluent_kafka.Consumer(config)
        topic_partitions_with_timestamp = [
            TopicPartition(topic.name, partition.partition_id, timestamp_limit.int_timestamp * 1000)
            for partition in topic.partitions
        ]
        topic_partitions_with_new_offsets = consumer.offsets_for_times(topic_partitions_with_timestamp)
        return {
            topic_partition.partition: topic_partition.offset for topic_partition in topic_partitions_with_new_offsets
        }

    def update_from_cluster(self, topic: Topic) -> Topic:
        """Takes a topic and, based on its name, updates all attributes from the cluster"""

        confluent_topic: ConfluentTopic = self._get_client_topic(topic.name, ClientTypes.Confluent)
        pykafka_topic: PyKafkaTopic = self._get_client_topic(topic.name, ClientTypes.PyKafka)
        low_watermarks = pykafka_topic.earliest_available_offsets()
        high_watermarks = pykafka_topic.latest_available_offsets()

        topic.partition_data = self._get_partition_data(confluent_topic, low_watermarks, high_watermarks, topic)
        topic.config = self.cluster.retrieve_config(ConfigResource.Type.TOPIC, topic.name)

        topic.is_only_local = False

        return topic

    def _get_partition_data(
        self,
        confluent_topic: ConfluentTopic,
        low_watermarks: PartitionInfo,
        high_watermarks: PartitionInfo,
        topic: Topic,
    ) -> List[Partition]:

        consumer = MessageConsumer(PING_GROUP_ID, topic.name, True)
        partitions = []

        for partition_id, meta in confluent_topic.partitions.items():
            low = int(low_watermarks[partition_id].offset[0])
            high = int(high_watermarks[partition_id].offset[0])
            latest_timestamp = None
            if high > low:
                try:
                    latest_timestamp = float(consumer.consume(high - 1, partition_id).timestamp()[1]) / 1000
                except MessageEmptyException:
                    logger.warning(
                        f"Due to timeout latest timestamp for topic `{topic.name}` and partition `{partition_id}` is missing."
                    )
            partition = Partition(partition_id, low, high, meta.isrs, meta.leader, meta.replicas, latest_timestamp)
            partitions.append(partition)

        return partitions

    def diff_with_cluster(self, local_topic: Topic) -> TopicDiff:
        assert local_topic.is_only_local, "Can only diff local topics with remote"

        cluster_topic = self.get_cluster_topic(local_topic.name)
        diffs = TopicDiff()
        diffs.set_diff("num_partitions", cluster_topic.num_partitions, local_topic.num_partitions)
        diffs.set_diff("replication_factor", cluster_topic.replication_factor, local_topic.replication_factor)

        for name, old_value in cluster_topic.config.items():
            diffs.set_diff(name, old_value, local_topic.config.get(name))

        return diffs



import pathlib
from abc import ABC, abstractmethod
from heapq import heappop, heappush
from typing import Dict, List, Optional, Tuple

import confluent_kafka
import pendulum
from confluent_kafka.cimpl import Message, TopicPartition

from esque.clients.schemaregistry import SchemaRegistryClient
from esque.config import Config
from esque.errors import EndOfPartitionReachedException, MessageEmptyException, raise_for_message
from esque.helpers import log_error
from esque.messages.avromessage import AvroFileWriter, StdOutAvroWriter
from esque.messages.message import FileWriter, GenericWriter, PlainTextFileWriter, StdOutWriter, decode_message
from esque.ruleparser.ruleengine import RuleTree

DEFAULT_CONSUME_TIMEOUT = 30  # seconds


class AbstractConsumer(ABC):
    def __init__(self, group_id: str, topic_name: str, last: bool, match: str = None, enable_auto_commit: bool = True):
        self._match = match
        self._last = last
        self._group_id = group_id
        self._consumer = None
        self._enable_auto_commit = enable_auto_commit
        self._config = {}
        if self._match is not None:
            self._rule_tree = RuleTree(match)
        else:
            self._rule_tree = None
        self._topic_name = topic_name
        self.writers: Dict[int, GenericWriter] = {}
        self._setup_config()
        self.create_internal_consumer()

    def _setup_config(self):
        offset_reset = "earliest"
        if self._last:
            offset_reset = "latest"
        self._config = Config.get_instance().create_confluent_config()
        self._config.update(
            {
                "group.id": self._group_id,
                "error_cb": log_error,
                # We need to commit offsets manually once we"re sure it got saved
                # to the sink
                "enable.auto.commit": self._enable_auto_commit,
                "enable.partition.eof": True,
                # We need this to start at the last committed offset instead of the
                # latest when subscribing for the first time
                "default.topic.config": {"auto.offset.reset": offset_reset},
            }
        )

    @abstractmethod
    def create_internal_consumer(self):
        raise NotImplementedError()

    def assign_specific_partitions(self, topic_name: str, partitions: list = None, offset: int = 0):
        self._topic_name = topic_name
        if partitions is not None:
            topic_partitions = [
                TopicPartition(self._topic_name, partition=partition, offset=offset) for partition in partitions
            ]
        else:
            topic_partitions = [TopicPartition(self._topic_name, partition=0, offset=offset)]
        self._consumer.assign(topic_partitions)

    def subscribe(self, topics: List[str]) -> None:
        self._consumer.subscribe(topics)

    def close(self) -> None:
        self._consumer.close()

    def commit(self, offsets: List[TopicPartition]):
        self._consumer.commit(offsets=offsets)

    @abstractmethod
    def consume(self, **kwargs) -> int:
        raise NotImplementedError()

    def output_consumed(self, message: Message):
        """
        Outputs the message to a destination determined by the implementation of the inheriting class.
        :param message: Message to output
        :return: This method returns no values
        """
        writer = self.writers.get(message.partition(), self.writers[-1])
        writer.write_message(message)

    def close_all_writers(self):
        for w in self.writers.values():
            if isinstance(w, FileWriter) and w.file is not None:
                w.file.close()

    def consume_single_message(self, timeout=DEFAULT_CONSUME_TIMEOUT) -> Message:
        message = self._consumer.poll(timeout=timeout)
        raise_for_message(message)
        return message

    def consume_single_acceptable_message(self, timeout=DEFAULT_CONSUME_TIMEOUT) -> Optional[Message]:
        message_acceptable = False
        total_time_remaining = timeout
        while not message_acceptable and total_time_remaining > 0:
            iteration_start = pendulum.now()
            message = self.consume_single_message(timeout=timeout)
            total_time_remaining -= (pendulum.now() - iteration_start).in_seconds()
            message_acceptable = self.consumed_message_matches(message)
        return message if message_acceptable else None

    def consumed_message_matches(self, message: Message):
        if self._rule_tree is not None:
            return self._rule_tree.evaluate(message)
        else:
            return True


class MessageConsumer(AbstractConsumer):
    def create_internal_consumer(self):
        self._consumer = confluent_kafka.Consumer(self._config)

    def consume(self, offset: int = 0, partition: int = 0) -> Message:
        self.assign_specific_partitions(self._topic_name, partitions=[partition], offset=offset)
        return self.consume_single_message()


class PingConsumer(AbstractConsumer):
    def consume(self) -> Optional[Tuple[str, int]]:
        message = self.consume_single_message()
        msg_sent_at = pendulum.from_timestamp(float(message.value()))
        delta_sent = pendulum.now() - msg_sent_at
        return message.key(), delta_sent.microseconds / 1000

    def create_internal_consumer(self):
        self._consumer = confluent_kafka.Consumer(self._config)
        self.assign_specific_partitions(self._topic_name, partitions=[0], offset=0)

    def _setup_config(self):
        super()._setup_config()
        self._config["enable.partition.eof"] = False


class PlaintextConsumer(AbstractConsumer):
    def __init__(
        self,
        group_id: str,
        topic_name: str,
        output_directory: pathlib.Path,
        last: bool,
        match: str = None,
        initialize_default_output_directory: bool = False,
        enable_auto_commit: bool = True,
    ):
        super().__init__(group_id, topic_name, last, match, enable_auto_commit)
        self.output_directory = output_directory
        self.writers[-1] = StdOutWriter() if output_directory is None else PlainTextFileWriter(self.output_directory)
        self._initialize_default_output_directory = initialize_default_output_directory
        if self._initialize_default_output_directory and self.output_directory is not None:
            self.writers[-1].init_destination_directory()

    def consume(self, amount: int) -> int:
        counter = 0

        while counter < amount:
            try:
                message = self.consume_single_acceptable_message()
            except MessageEmptyException:
                return counter
            except EndOfPartitionReachedException:
                pass
            else:
                counter += 1
                self.output_consumed(message)

        self.close_all_writers()
        return counter

    def create_internal_consumer(self):
        self._consumer = confluent_kafka.Consumer(self._config)
        if self._topic_name is not None:
            self.subscribe([self._topic_name])

    def output_consumed(self, message: Message):
        if (
            self.output_directory
            and not self._initialize_default_output_directory
            and message.partition() not in self.writers
        ):
            writer = PlainTextFileWriter(self.output_directory, message.partition())
            writer.init_destination_directory()
            self.writers[message.partition()] = writer
        else:
            writer = self.writers.get(message.partition(), self.writers[-1])
        writer.write_message(message)


class AvroFileConsumer(PlaintextConsumer):
    def __init__(
        self,
        group_id: str,
        topic_name: str,
        output_directory: pathlib.Path,
        last: bool,
        match: str = None,
        initialize_default_output_directory: bool = False,
        enable_auto_commit: bool = True,
    ):
        super().__init__(
            group_id,
            topic_name,
            output_directory,
            last,
            match,
            initialize_default_output_directory,
            enable_auto_commit,
        )
        self.schema_registry_client = SchemaRegistryClient(Config.get_instance().schema_registry)
        self.writers[-1] = (
            StdOutAvroWriter(schema_registry_client=self.schema_registry_client)
            if output_directory is None
            else AvroFileWriter(self.schema_registry_client, self.output_directory)
        )
        if self._initialize_default_output_directory and self.output_directory is not None:
            self.writers[-1].init_destination_directory()

    def output_consumed(self, message: Message):
        if (
            self.output_directory
            and not self._initialize_default_output_directory
            and message.partition() not in self.writers
        ):
            writer = AvroFileWriter(self.schema_registry_client, self.output_directory, message.partition())
            writer.init_destination_directory()
            self.writers[message.partition()] = writer
        else:
            writer = self.writers.get(message.partition(), self.writers[-1])

        writer.write_message(message)


class ConsumerFactory:
    def create_consumer(
        self,
        group_id: str,
        topic_name: str,
        output_directory: pathlib.Path,
        last: bool,
        avro: bool,
        initialize_default_output_directory: bool = False,
        match: str = None,
        enable_auto_commit: bool = True,
    ):
        """
        Creates a Kafka consumer
        :param group_id: ID of the consumer group
        :param topic_name: Topic name for the new consumer
        :param output_directory: The directory to store the consumed messages (if None, than STDOUT is used)
        :param last: Start consuming from the latest committed offset
        :param avro: Are messages in Avro format?
        :param initialize_default_output_directory: If set to true, all messages will be stored in a directory named partition_any, instead having a separate directory for each partition. This argument is only used if output_directory is not None.
        :param match: Match expression for message filtering
        :param enable_auto_commit: Allow the consumer to automatically commit offset
        :return: Consumer object
        """
        if avro:
            consumer = AvroFileConsumer(
                group_id=group_id,
                topic_name=topic_name,
                output_directory=output_directory,
                last=last,
                match=match,
                initialize_default_output_directory=initialize_default_output_directory,
                enable_auto_commit=enable_auto_commit,
            )
        else:
            consumer = PlaintextConsumer(
                group_id=group_id,
                topic_name=topic_name,
                output_directory=output_directory,
                last=last,
                match=match,
                initialize_default_output_directory=initialize_default_output_directory,
                enable_auto_commit=enable_auto_commit,
            )
        return consumer

    def create_ping_consumer(self, group_id: str, topic_name: str):
        consumer = PingConsumer(group_id, topic_name, last=False)
        return consumer


def consume_to_file_ordered(
    output_directory: pathlib.Path,
    topic: str,
    group_id: str,
    partitions: list,
    numbers: int,
    avro: bool,
    match: str,
    last: bool,
    write_to_stdout: bool = False,
) -> int:
    consumers = []
    factory = ConsumerFactory()
    for partition in partitions:
        consumer = factory.create_consumer(
            group_id=group_id + "_" + str(partition),
            topic_name=None,
            output_directory=None if write_to_stdout else output_directory,
            avro=avro,
            match=match,
            last=last,
            initialize_default_output_directory=True,
        )
        consumer.assign_specific_partitions(topic, [partition])
        consumers.append(consumer)

    message_heap = []
    total_number_of_messages = 0
    messages_left = True
    # get at least one message from each partition, or exclude those that don't have any messages
    for partition_counter in range(0, len(consumers)):
        max_retry_count = 5
        keep_polling_current_partition = True
        while keep_polling_current_partition:
            try:
                message = consumers[partition_counter].consume_single_acceptable_message()
                decoded_message = decode_message(message)
            except MessageEmptyException:
                # a possible timeout due to a network issue, retry (but not more than max_retry_count attempts)
                max_retry_count -= 1
                if max_retry_count <= 0:
                    partitions.remove(partition_counter)
                    if len(partitions) == 0:
                        messages_left = False
                    keep_polling_current_partition = False
            except EndOfPartitionReachedException:
                keep_polling_current_partition = False
                partitions.remove(partition_counter)
                if len(partitions) == 0:
                    messages_left = False
            else:
                keep_polling_current_partition = False
                heappush(message_heap, (decoded_message.timestamp, message))

    # in each iteration, take the earliest message from the map, output it and replace it with a new one (if available)
    # if not, remove the consumer and move to the next one
    while total_number_of_messages < numbers and messages_left:
        if len(message_heap) == 0:
            messages_left = False
        else:
            (timestamp, message) = heappop(message_heap)
            consumers[0].output_consumed(message)
            total_number_of_messages += 1
            partition = message.partition()
            try:
                message = consumers[partition].consume_single_acceptable_message()
                decoded_message = decode_message(message)
                heappush(message_heap, (decoded_message.timestamp, message))
            except (MessageEmptyException, EndOfPartitionReachedException):
                partitions.remove(partition)
                if len(partitions) == 0:
                    messages_left = False
    for c in consumers:
        c.close_all_writers()
    return total_number_of_messages


def consume_to_files(
    output_directory: pathlib.Path,
    topic: str,
    group_id: str,
    numbers: int,
    avro: bool,
    match: str,
    last: bool,
    write_to_stdout: bool = False,
) -> int:
    consumer = ConsumerFactory().create_consumer(
        group_id=group_id,
        topic_name=topic,
        output_directory=output_directory if not write_to_stdout else None,
        last=last,
        avro=avro,
        match=match,
        initialize_default_output_directory=False,
    )
    number_consumed_messages = consumer.consume(int(numbers))
    consumer.close_all_writers()
    return number_consumed_messages



import functools
import logging
from concurrent.futures import Future, wait
from itertools import islice
from typing import Type, TypeVar

import confluent_kafka
import pendulum
from confluent_kafka.cimpl import KafkaError

from esque.errors import FutureTimeoutException, raise_for_kafka_error

log = logging.getLogger(__name__)

T = TypeVar("T")


class SingletonMeta(type):
    __instance: T

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        cls.__instance = None

    def get_instance(cls: Type[T]) -> T:
        if cls.__instance is None:
            cls.set_instance(cls())
        return cls.__instance

    def set_instance(cls: Type[T], instance: T):
        cls.__instance = instance


def invalidate_cache_before(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        self._client.poll(timeout=1)
        return func(self, *args, **kwargs)

    return wrapper


def invalidate_cache_after(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        result = func(self, *args, **kwargs)
        self.cluster.confluent_client.poll(timeout=1)
        return result

    return wrapper


def ensure_kafka_future_done(future: Future, timeout: int = 60 * 5) -> Future:
    # Clients, such as confluents AdminClient, may return a done future with an exception
    done, not_done = wait({future}, timeout=timeout)

    if not_done:
        raise FutureTimeoutException("Future timed out after {} seconds".format(timeout))

    result = next(islice(done, 1))

    exception = result.exception()

    if exception is None:
        return result
    elif isinstance(exception, confluent_kafka.KafkaException):
        raise_for_kafka_error(exception.args[0])
    else:
        raise exception


def unpack_confluent_config(config):
    return {v.name: v.value for k, v in sorted(config.items())}


def log_error(err: KafkaError):
    if err is not None:
        log.error(f"KafkaError occured: {err.str()}")


def delta_t(start: pendulum.DateTime) -> str:
    now = pendulum.now()
    return now.to_datetime_string() + " " + (pendulum.now() - start).in_words()



from abc import ABC
from enum import Enum
from typing import Callable, List, NamedTuple

import pendulum

import esque.ruleparser.helpers as h


class OperatorType(Enum):
    ARITHMETIC = (1,)
    COMPARISON = (2,)
    PARENTHESIS = 3


class OperandType(Enum):
    NUMERIC = (1,)
    NUMERIC_INT = (2,)
    BOOLEAN = (3,)
    DATETIME = (4,)
    STRING = 5


class GenericOperator(NamedTuple):
    name: str
    type: OperatorType
    regex: str
    input_priority: int
    stack_priority: int
    is_unary: bool
    calculation_function: Callable
    operand_types: List[OperandType]


class Operator(ABC):
    generic_operator: GenericOperator = None
    literal: str = ""

    def __init__(self, literal: str, generic_operator: GenericOperator):
        self.literal = literal.replace("\\", "")
        self.generic_operator = generic_operator

    FIELDS = {
        "MESSAGE_KEY": "message\\.key",
        "MESSAGE_LENGTH": "message\\.length",
        "MESSAGE_PARTITION": "message\\.partition",
        "MESSAGE_HEADER": "message\\.header\\.[a-zA-Z0-9_]+",
        "MESSAGE_OFFSET": "message\\.offset",
        "MESSAGE_TIMESTAMP": "message\\.timestamp",
        "SYSTEM_TIMESTAMP": "system\\.timestamp",
    }
    OPERATORS = {
        "ADDITION": GenericOperator(
            "ADDITION", OperatorType.ARITHMETIC, "\\+", 20, 20, False, (lambda o1, o2: o1 + o2), [OperandType.NUMERIC]
        ),
        "SUBTRACTION": GenericOperator(
            "SUBTRACTION", OperatorType.ARITHMETIC, "-", 20, 20, False, (lambda o1, o2: o1 - o2), [OperandType.NUMERIC]
        ),
        "MULTIPLICATION": GenericOperator(
            "MULTIPLICATION",
            OperatorType.ARITHMETIC,
            "\\*",
            30,
            30,
            False,
            (lambda o1, o2: o1 * o2),
            [OperandType.NUMERIC],
        ),
        "DIVISION": GenericOperator(
            "DIVISION", OperatorType.ARITHMETIC, "\\/", 30, 30, False, (lambda o1, o2: o1 / o2), [OperandType.NUMERIC]
        ),
        "REMAINDER": GenericOperator(
            "REMAINDER",
            OperatorType.ARITHMETIC,
            "mod",
            20,
            20,
            False,
            (lambda o1, o2: o1 % o2),
            [OperandType.NUMERIC_INT],
        ),
        "BINARY_AND": GenericOperator(
            "BINARY_AND",
            OperatorType.ARITHMETIC,
            "&",
            20,
            20,
            False,
            (lambda o1, o2: o1 & o2),
            [OperandType.NUMERIC_INT],
        ),
        "BINARY_OR": GenericOperator(
            "BINARY_OR",
            OperatorType.ARITHMETIC,
            "\\|",
            20,
            20,
            False,
            (lambda o1, o2: o1 | o2),
            [OperandType.NUMERIC_INT],
        ),
        "BINARY_XOR": GenericOperator(
            "BINARY_XOR",
            OperatorType.ARITHMETIC,
            "\\^",
            20,
            20,
            False,
            (lambda o1, o2: o1 ^ o2),
            [OperandType.NUMERIC_INT],
        ),
        "GREATER_THAN": GenericOperator(
            "GREATER_THAN",
            OperatorType.COMPARISON,
            ">",
            15,
            15,
            False,
            (lambda o1, o2: o1 > o2),
            [OperandType.NUMERIC, OperandType.DATETIME],
        ),
        "LESS_THAN": GenericOperator(
            "LESS_THAN",
            OperatorType.COMPARISON,
            "<",
            15,
            15,
            False,
            (lambda o1, o2: o1 < o2),
            [OperandType.NUMERIC, OperandType.DATETIME],
        ),
        "GREATER_OR_EQUAL": GenericOperator(
            "GREATER_OR_EQUAL",
            OperatorType.COMPARISON,
            ">=",
            15,
            15,
            False,
            (lambda o1, o2: o1 >= o2),
            [OperandType.NUMERIC, OperandType.DATETIME],
        ),
        "LESS_OR_EQUAL": GenericOperator(
            "LESS_OR_EQUAL",
            OperatorType.COMPARISON,
            "<=",
            15,
            15,
            False,
            (lambda o1, o2: o1 <= o2),
            [OperandType.NUMERIC, OperandType.DATETIME],
        ),
        "EQUAL": GenericOperator(
            "EQUAL",
            OperatorType.COMPARISON,
            "==",
            15,
            15,
            False,
            (lambda o1, o2: o1 == o2),
            [OperandType.NUMERIC, OperandType.DATETIME, OperandType.BOOLEAN, OperandType.STRING],
        ),
        "NOT_EQUAL": GenericOperator(
            "NOT_EQUAL",
            OperatorType.COMPARISON,
            "!=",
            15,
            15,
            False,
            (lambda o1, o2: o1 != o2),
            [OperandType.NUMERIC, OperandType.DATETIME, OperandType.BOOLEAN, OperandType.STRING],
        ),
        "LOGICAL_AND": GenericOperator(
            "LOGICAL_AND",
            OperatorType.ARITHMETIC,
            "and",
            10,
            10,
            False,
            (lambda o1, o2: o1 and o2),
            [OperandType.BOOLEAN],
        ),
        "LOGICAL_OR": GenericOperator(
            "LOGICAL_OR",
            OperatorType.ARITHMETIC,
            "or",
            10,
            10,
            False,
            (lambda o1, o2: o1 or o2),
            [OperandType.BOOLEAN],
        ),
        "LOGICAL_NOT": GenericOperator(
            "LOGICAL_NOT", OperatorType.COMPARISON, "neg", 100, 100, True, (lambda o1: not o1), [OperandType.BOOLEAN]
        ),
        "LOGICAL_XOR": GenericOperator(
            "LOGICAL_XOR",
            OperatorType.ARITHMETIC,
            "xor",
            10,
            10,
            False,
            (lambda o1, o2: (o1 and not o2) or (not o1 and o2)),
            [OperandType.BOOLEAN],
        ),
        "LIKE": GenericOperator(
            "LIKE",
            OperatorType.COMPARISON,
            "like",
            15,
            15,
            False,
            (lambda o1, o2: h.string_like(haystack=o1, needle=o2)),
            [OperandType.STRING],
        ),
        "NOT_LIKE": GenericOperator(
            "NOT_LIKE",
            OperatorType.COMPARISON,
            "notlike",
            15,
            15,
            False,
            (lambda o1, o2: h.string_not_like(haystack=o1, needle=o2)),
            [OperandType.STRING],
        ),
        "PARENTHESIS_OPEN": GenericOperator(
            "PARENTHESIS_OPEN", OperatorType.PARENTHESIS, "\\(", 10000, 0, False, None, None
        ),
        "PARENTHESIS_CLOSED": GenericOperator(
            "PARENTHESIS_CLOSED", OperatorType.PARENTHESIS, "\\)", 1, 0, False, None, None
        ),
    }


class AbstractBinaryOperator(Operator):
    def evaluate(self, operand1, operand2):
        pass

    def validate_operands(self, operand1, operand2):
        for operand_type in self.generic_operator.operand_types:
            if (
                (operand_type == OperandType.NUMERIC_INT or operand_type == OperandType.NUMERIC)
                and h.is_int(operand1)
                and h.is_int(operand2)
            ):
                return [int(operand1), int(operand2)]
            elif operand_type == OperandType.NUMERIC and h.is_float(operand1) and h.is_float(operand2):
                return [float(operand1), float(operand2)]
            elif (
                operand_type == OperandType.DATETIME
                and (h.is_date_time_string(operand1) or h.is_date_string(operand1))
                and (h.is_date_time_string(operand2) or h.is_date_string(operand2))
            ):
                return [pendulum.parse(operand1), pendulum.parse(operand2)]
            elif operand_type == OperandType.BOOLEAN and h.is_boolean(operand1) and h.is_boolean(operand2):
                return [h.to_boolean(operand1), h.to_boolean(operand2)]
            elif operand_type == OperandType.STRING:
                return [operand1, operand2]
        raise ValueError(
            "One or both operands ({}, {}) are not applicable to this operator ({})".format(
                operand1, operand2, self.literal
            )
        )


class AbstractUnaryOperator(Operator):
    def evaluate(self, operand1):
        pass

    def validate_operands(self, operand1):
        for operand_type in self.generic_operator.operand_types:
            if (operand_type == OperandType.NUMERIC_INT or operand_type == OperandType.NUMERIC) and h.is_int(operand1):
                return [int(operand1)]
            elif operand_type == OperandType.NUMERIC and h.is_float(operand1):
                return [float(operand1)]
            elif operand_type == OperandType.DATETIME and (
                h.is_date_time_string(operand1) or h.is_date_string(operand1)
            ):
                return [pendulum.parse(operand1)]
            elif operand_type == OperandType.BOOLEAN and h.is_boolean(operand1):
                return [h.to_boolean(operand1)]
            elif operand_type == OperandType.STRING:
                return [operand1]
        raise ValueError("Operand ({}) is not applicable to this operator ({})".format(operand1, self.literal))


class ArithmeticBinaryOperator(AbstractBinaryOperator):
    def evaluate(self, operand1: str, operand2: str):
        try:
            [op1_converted, op2_converted] = self.validate_operands(operand1, operand2)
            return self.generic_operator.calculation_function(op1_converted, op2_converted)
        except ValueError:
            raise


class ComparisonBinaryOperator(AbstractBinaryOperator):
    def evaluate(self, operand1: str, operand2: str):
        try:
            [op1_converted, op2_converted] = self.validate_operands(operand1, operand2)
            return self.generic_operator.calculation_function(op1_converted, op2_converted)
        except ValueError:
            raise


class ComparisonUnaryOperator(AbstractUnaryOperator):
    def evaluate(self, operand1: str):
        try:
            [op1_converted] = self.validate_operands(operand1)
            return self.generic_operator.calculation_function(op1_converted)
        except ValueError:
            raise


class ParenthesisOperator(Operator):
    pass


class Literal(ABC):
    def __init__(self, value: str):
        self.value = value.strip()


class Field(ABC):
    def __init__(self, field_name: str):
        self.field_name = field_name.strip()



from typing import Dict, Optional, Type

import pykafka.exceptions
from confluent_kafka import KafkaError, Message


def raise_for_kafka_error(err: KafkaError, msg: Optional[Message] = None):
    if not err:
        return None

    if err.code() in CONFLUENT_ERROR_LOOKUP.keys():
        raise CONFLUENT_ERROR_LOOKUP[err.code()](err.str(), err.code())
    else:
        raise KafkaException(err.str(), err.code())


def raise_for_message(message: Message):
    if message is None:
        raise MessageEmptyException
    elif message.error() is not None:
        raise_for_kafka_error(message.error(), message)


class EsqueException(Exception):
    pass


class ExceptionWithMessage(EsqueException):
    def __init__(self, message):
        self.message = message

    def format_message(self):
        return f"{type(self).__name__}: {self.message}"


class EditCanceled(ExceptionWithMessage):
    def __init__(self):
        super().__init__("Edit canceled.")


class ConfigVersionException(ExceptionWithMessage):
    def __init__(self, current, expected):
        super().__init__(f"Expected config version {expected}, got {current}.")


class ConfigTooOld(ConfigVersionException):
    def __init__(self, current, expected):
        super().__init__(current, expected)
        self.message += "\nRun `esque config migrate` to migrate to new version."


class ConfigTooNew(ConfigVersionException):
    def __init__(self, current, expected):
        super().__init__(current, expected)
        self.message += "\nYou might find a backup from last config migration next to your current config."


class ConfigException(ExceptionWithMessage):
    pass


class MissingSaslParameter(ConfigException):
    pass


class UnsupportedSaslMechanism(ConfigException):
    pass


class KafkaException(ExceptionWithMessage):
    def __init__(self, message: str, code: int):
        super().__init__(f"{message} with code {code}.")
        self.code = code


class ConsumerGroupDoesNotExistException(ExceptionWithMessage):
    def __init__(self, consumer_id: str):
        super().__init__(f"Consumer Group does not exist for consumer id '{self.consumer_id}'.")
        self.consumer_id = consumer_id


class ConfigNotExistsException(ExceptionWithMessage):
    def __init__(self):
        super().__init__("Config does not exist.")


class NoConfirmationPossibleException(ExceptionWithMessage):
    def __init__(self):
        super().__init__(
            "You are running this command in a non-interactive mode. To do this you must use the --no-verify option."
        )


class ContextNotDefinedException(ExceptionWithMessage):
    def __init__(self, message: str = None):
        if message is None:
            message = "Context cannot be found."
        super().__init__(message)


class FutureTimeoutException(ExceptionWithMessage):
    pass


class MessageEmptyException(KafkaException):
    def __init__(self):
        super().__init__("Consumed Message is empty.", -185)


class TopicAlreadyExistsException(KafkaException):
    pass


class EndOfPartitionReachedException(KafkaException):
    pass


class TopicDoesNotExistException(KafkaException):
    pass


class InvalidReplicationFactorException(KafkaException):
    pass


class ConnectionFailedException(ExceptionWithMessage):
    def __init__(self, pykafka_exception: pykafka.exceptions.KafkaException):
        self.pykafka_exception = pykafka_exception

        if isinstance(self.pykafka_exception.args, str):
            msg = self.pykafka_exception.args
        else:
            msg = f"Connection to brokers failed."
        super().__init__(msg)


class ValidationException(ExceptionWithMessage):
    pass


class YamaleValidationException(ValidationException):
    def __init__(self, validation_error: ValueError):
        complete_message = validation_error.args[0]
        messages = complete_message.split("\n")[2:]
        stripped_messages = list(map(lambda x: x.strip("\t"), messages))
        joined_message = "\n".join(stripped_messages)
        super().__init__(joined_message)


CONFLUENT_ERROR_LOOKUP: Dict[int, Type[KafkaException]] = {
    KafkaError.UNKNOWN_TOPIC_OR_PART: TopicDoesNotExistException,
    KafkaError.TOPIC_ALREADY_EXISTS: TopicAlreadyExistsException,
    KafkaError._PARTITION_EOF: EndOfPartitionReachedException,
    KafkaError.INVALID_REPLICATION_FACTOR: InvalidReplicationFactorException,
}



import operator

import pykafka
from confluent_kafka.admin import AdminClient, ConfigResource

from esque.config import Config
from esque.controller.topic_controller import TopicController
from esque.helpers import ensure_kafka_future_done, unpack_confluent_config


class Cluster:
    def __init__(self):
        self._config = Config.get_instance()
        self.confluent_client = AdminClient(self._config.create_confluent_config())
        self.pykafka_client = pykafka.client.KafkaClient(
            **self._config.create_pykafka_config(), broker_version="1.0.0"
        )
        self.confluent_client.poll(timeout=1)
        self.__topic_controller = None

    @property
    def topic_controller(self) -> TopicController:
        if self.__topic_controller is None:
            self.__topic_controller = TopicController(self, self._config)
        return self.__topic_controller

    @property
    def bootstrap_servers(self):
        return self._config.bootstrap_servers

    def get_metadata(self):
        return self.confluent_client.list_topics(timeout=1)

    @property
    def brokers(self):
        metadata = self.confluent_client.list_topics(timeout=1)
        return sorted(
            [{"id": broker.id, "host": broker.host, "port": broker.port} for broker in metadata.brokers.values()],
            key=operator.itemgetter("id"),
        )

    def retrieve_config(self, config_type: ConfigResource.Type, id):
        requested_resources = [ConfigResource(config_type, str(id))]
        futures = self.confluent_client.describe_configs(requested_resources)
        (old_resource, future), = futures.items()
        future = ensure_kafka_future_done(future)
        result = future.result()
        return unpack_confluent_config(result)



import datetime
import re
from typing import Any

from confluent_kafka.cimpl import Message

from esque.messages.message import KafkaMessage, decode_message
from esque.ruleparser.expressionelement import Operator


class FieldEval:

    __message = None
    __header_pattern = None
    __time_format: str = "%Y-%m-%dT%H:%M:%S"

    def __init__(self, message: Any = None):
        self.__message = message
        self.__header_pattern = re.compile(Operator.FIELDS["MESSAGE_HEADER"], flags=re.IGNORECASE)

    def evaluate_field(self, field_name: str):
        """ This method contains the logic to evaluate the field values. New fields are added as follows:
            (1) add a new element to the Operator.FIELDS set. The key is arbitrary, and the field contains a regex for the parser,
            (2) add an implementation to this method
        """
        if field_name == Operator.FIELDS["SYSTEM_TIMESTAMP"].replace("\\", ""):
            return datetime.datetime.strftime(datetime.datetime.now(), self.__time_format)
        elif self.__message is None:
            return -1
        elif isinstance(self.__message, Message):
            return self.__evaluate_message_field(field_name=field_name)
        elif isinstance(self.__message, KafkaMessage):
            return self.__evaluate_kafka_message_field(field_name=field_name)

    def __evaluate_message_field(self, field_name: str):
        if field_name == Operator.FIELDS["MESSAGE_OFFSET"].replace("\\", ""):
            return self.__message.offset()
        elif field_name == Operator.FIELDS["MESSAGE_PARTITION"].replace("\\", ""):
            partition = self.__message.partition()
            return -1 if not partition else partition
        elif field_name == Operator.FIELDS["MESSAGE_TIMESTAMP"].replace("\\", ""):
            return datetime.datetime.strftime(
                datetime.datetime.fromtimestamp(self.__message.timestamp()[1] / 1000.0), self.__time_format
            )
        elif field_name == Operator.FIELDS["MESSAGE_LENGTH"].replace("\\", ""):
            return len(self.__message)
        elif field_name == Operator.FIELDS["MESSAGE_KEY"].replace("\\", ""):
            return self.__message.key().decode("UTF-8")
        elif self.__header_pattern.fullmatch(field_name):
            header_name = field_name.split(".")[-1]
            if self.__message.headers() is not None:
                message_headers = decode_message(message=self.__message).headers
                return next(
                    (message_header.value for message_header in message_headers if message_header.key == header_name),
                    "no_header",
                )
            return "no_header"
        return -1

    def __evaluate_kafka_message_field(self, field_name: str):
        if field_name == Operator.FIELDS["MESSAGE_OFFSET"].replace("\\", ""):
            return -1
        elif field_name == Operator.FIELDS["MESSAGE_PARTITION"].replace("\\", ""):
            partition = self.__message.partition
            return -1 if not partition else partition
        elif field_name == Operator.FIELDS["MESSAGE_LENGTH"].replace("\\", ""):
            return len(self.__message.key) + len(self.__message.value)
        elif field_name == Operator.FIELDS["MESSAGE_KEY"].replace("\\", ""):
            return self.__message.key
        elif self.__header_pattern.fullmatch(field_name):
            header_name = field_name.split(".")[-1]
            message_headers = self.__message.headers
            return next(
                (message_header.value for message_header in message_headers if message_header.key == header_name),
                "no_header",
            )
        return -1



import operator
from typing import Dict, List

from confluent_kafka.admin import ConfigResource

from esque.resources.resource import KafkaResource


class Broker(KafkaResource):
    def __init__(self, cluster, *, broker_id: int = None, host: str = None, port: int = None):
        self.cluster = cluster
        self.broker_id = broker_id
        self.host = host
        self.port = port

    @classmethod
    def from_id(cls, cluster, broker_id) -> "Broker":
        return cls(cluster=cluster, broker_id=broker_id)

    @classmethod
    def from_host(cls, cluster, host: str) -> "Broker":
        brokers = [broker for broker in cluster.brokers if broker["host"] == host]
        if len(brokers) > 1:
            raise ValueError(
                f"Broker host name {host} is not unique! Please provide with port number i.e. {host}:port."
            )
        elif len(brokers) == 0:
            raise ValueError(f"There is no broker with {host} as host name!")
        else:
            broker = brokers[0]
            return cls(cluster, broker_id=broker["id"], host=host, port=broker["port"])

    @classmethod
    def from_host_and_port(cls, cluster, host: str, port: int) -> "Broker":
        brokers = cluster.brokers
        for broker in brokers:
            if broker["host"] == host and broker["port"] == port:
                return cls(cluster, broker_id=broker["id"], host=host, port=port)

    @classmethod
    def from_attributes(cls, cluster, broker_id: int, host: str, port: int) -> "Broker":
        return cls(cluster, broker_id=broker_id, host=host, port=port)

    @classmethod
    def get_all(cls, cluster) -> List["Broker"]:
        metadata = cluster.get_metadata().brokers.values()
        brokers = [
            cls.from_attributes(cluster, broker_id=broker.id, host=broker.host, port=broker.port)
            for broker in metadata
        ]
        return sorted(brokers, key=operator.attrgetter("broker_id"))

    def describe(self) -> Dict:
        return self.cluster.retrieve_config(ConfigResource.Type.BROKER, self.broker_id)

    def as_dict(self):
        return {"cluster": self.cluster, "broker_id": self.broker_id, "host": self.host, "port": self.port}



import sys
from functools import wraps
from shutil import copyfile

import click
from click import make_pass_decorator, option

from esque.cli import environment
from esque.cli.helpers import ensure_approval
from esque.cluster import Cluster
from esque.config import Config, config_dir, config_path, sample_config_path
from esque.errors import ConfigNotExistsException


class State(object):
    def __init__(self):
        self.no_verify = False
        self._verbose = False
        self._cluster = None
        self._config = None

    @property
    def config(self) -> Config:
        if self._config is None:
            self._create_config()
        return self._config

    def _create_config(self):
        try:
            self._config = Config.get_instance()
        except ConfigNotExistsException:
            click.echo(f"No config provided in {config_dir()}")
            if ensure_approval(f"Should a sample file be created in {config_dir()}"):
                config_dir().mkdir(exist_ok=True)
                copyfile(sample_config_path().as_posix(), config_path())
            else:
                raise
            if ensure_approval("Do you want to modify the config file now?"):
                click.edit(filename=config_path().as_posix())
            self._config = Config.get_instance()

    @property
    def cluster(self):
        if not self._cluster:
            self._cluster = Cluster()
        return self._cluster

    def _get_verbose(self) -> bool:
        if environment.ESQUE_VERBOSE is not None:
            return True
        return self._verbose

    def _set_verbose(self, verbose):
        self._verbose = verbose

    verbose = property(_get_verbose, _set_verbose)


pass_state = make_pass_decorator(State, ensure=True)


def verbose_callback(context, _: str, verbose=False):
    state = context.ensure_object(State)
    state.verbose = verbose


verbose_option = click.option(
    "-v",
    "--verbose",
    is_flag=True,
    is_eager=True,
    callback=verbose_callback,
    expose_value=False,
    help="Return stack trace on error.",
)


def default_options(f):
    defaults = [no_verify_option, verbose_option, error_handler, pass_state]
    for decorator in defaults:
        f = decorator(f)
    return f


def no_verify_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(State)
        state.no_verify = value

    return option(
        "--no-verify",
        type=bool,
        help="Skip all verification dialogs and answer them with yes.",
        required=False,
        is_flag=True,
        expose_value=False,
        default=False,
        callback=callback,
    )(f)


output_format_option = click.option(
    "-o",
    "--output-format",
    type=click.Choice(["yaml", "json"], case_sensitive=False),
    help="Format of the output.",
    required=False,
)


def error_handler(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        state = args[0]
        if not isinstance(state, State):
            raise TypeError(
                "First argument is not a state, make sure that the `error_handler` decorator comes below `pass_state`"
            )
        try:
            f(*args, **kwargs)
        except Exception as e:
            if state.verbose:
                raise
            _silence_exception(e)

    return wrapper


def _silence_exception(e: Exception):
    if hasattr(e, "format_message"):
        click.echo(e.format_message())
    elif isinstance(e, (KeyError, ValueError)):
        click.echo(f"{type(e).__name__}: {str(e)}")
    else:
        click.echo(f"Exception of type {type(e).__name__} occurred.")
    click.echo("Run with `--verbose` for complete error.")
    sys.exit(1)



import getpass
import logging
import pwd
import sys
import time
from itertools import groupby
from pathlib import Path
from shutil import copyfile
from time import sleep

import click
import yaml
from click import MissingParameter, version_option

from esque import __version__, validation
from esque.cli.helpers import attrgetter, edit_yaml, ensure_approval, isatty
from esque.cli.options import State, default_options, output_format_option
from esque.cli.output import (
    blue_bold,
    bold,
    format_output,
    green_bold,
    pretty,
    pretty_new_topic_configs,
    pretty_topic_diffs,
    pretty_unchanged_topic_configs,
    red_bold,
)
from esque.clients.consumer import ConsumerFactory, consume_to_file_ordered, consume_to_files
from esque.clients.producer import PingProducer, ProducerFactory
from esque.config import PING_GROUP_ID, PING_TOPIC, Config, config_dir, config_path, migration, sample_config_path
from esque.controller.consumergroup_controller import ConsumerGroupController
from esque.errors import TopicAlreadyExistsException, TopicDoesNotExistException, ValidationException
from esque.resources.broker import Broker
from esque.resources.topic import Topic, copy_to_local


@click.group(invoke_without_command=True, no_args_is_help=True)
@version_option(__version__)
@default_options
def esque(state: State):
    """esque - an operational kafka tool.

    In the Kafka world nothing is easy, but esque (pronounced esk) is an attempt at it.
    """
    pass


@esque.group(help="Get a quick overview of different resources.", no_args_is_help=True)
@default_options
def get(state: State):
    pass


@esque.group(help="Get detailed information about a resource.", no_args_is_help=True)
@default_options
def describe(state: State):
    pass


@esque.group(help="Create a new instance of a resource.", no_args_is_help=True)
@default_options
def create(state: State):
    pass


@esque.group(help="Delete a resource.", no_args_is_help=True)
@default_options
def delete(state: State):
    pass


@esque.group(help="Edit a resource.", no_args_is_help=True)
@default_options
def edit(state: State):
    pass


@esque.group(help="Configuration-related options.", no_args_is_help=True)
@default_options
def config(state: State):
    pass


@esque.group(name="set", help="Set resource attributes.")
@default_options
def set_(state: State):
    pass


def list_brokers(ctx, args, incomplete):
    state = ctx.ensure_object(State)
    all_broker_hosts_names = [f"{broker.host}:{broker.port}" for broker in Broker.get_all(state.cluster)]
    return [broker for broker in all_broker_hosts_names if broker.startswith(incomplete)]


def list_consumergroups(ctx, args, incomplete):
    state = ctx.ensure_object(State)
    return [
        group
        for group in ConsumerGroupController(state.cluster).list_consumer_groups()
        if group.startswith(incomplete)
    ]


def list_contexts(ctx, args, incomplete):
    state = ctx.ensure_object(State)
    return [context for context in state.config.available_contexts if context.startswith(incomplete)]


def list_topics(ctx, args, incomplete):
    state = ctx.ensure_object(State)
    cluster = state.cluster
    return [
        topic.name for topic in cluster.topic_controller.list_topics(search_string=incomplete, get_topic_objects=False)
    ]


def fallback_to_stdin(ctx, args, value):
    stdin = click.get_text_stream("stdin")
    if not value and not isatty(stdin):
        stdin_arg = stdin.readline().strip()
    else:
        stdin_arg = value
    if not stdin_arg:
        raise MissingParameter("No value specified!")

    return stdin_arg


@esque.command("ctx")
@click.argument("context", required=False, default=None, autocompletion=list_contexts)
@default_options
def ctx(state: State, context: str):
    """List contexts and switch between them.

    \b
    USAGE:
    esque ctx               : list available contexts
    esque ctx CONTEXT       : switch to context CONTEXT
    """
    if not context:
        for c in state.config.available_contexts:
            if c == state.config.current_context:
                click.echo(bold(c))
            else:
                click.echo(c)
    if context:
        state.config.context_switch(context)
        state.config.save()
        click.echo(f"Switched to context: {context}.")


@config.command("recreate")
@default_options
def config_recreate(state: State):
    """(Re)create esque config.

    Overwrites the existing esque config file with the sample config. If no esque config file already exists,
    create one with the sample config."""
    config_dir().mkdir(exist_ok=True)
    if ensure_approval(f"Should the current config in {config_dir()} get replaced?", no_verify=state.no_verify):
        copyfile(sample_config_path().as_posix(), config_path())


@config.command("fix")
@default_options
def config_fix(state: State):
    """Fix simple errors in esque config.

    Fixes simple errors like wrong current_contexts in the esque config when the configs was tampered with manually."""
    try:
        state.config.context_switch(state.config.current_context)
        click.echo("Your config seems fine. 🎉")
    except ValidationException:
        _cfg: Config = Config(disable_validation=True)
        if _cfg.current_context not in _cfg.available_contexts:
            click.echo(f"Found invalid current context. Switching context to state {_cfg.available_contexts[0]}.")
            _cfg.context_switch(_cfg.available_contexts[0])
            Config.set_instance(_cfg)
            state.config.save()
        else:
            click.echo("Can't fix this configuration error try fixing it manually.")


@config.command("autocomplete")
@default_options
def config_autocomplete(state: State):
    """Configure esque autocompletion functionality.

    Generate the autocompletion script based on the current shell and
    give instructions to install it into the current environment.
    """
    directory = config_dir()
    config_file_name = "autocomplete.sh"
    config_file: Path = directory / config_file_name
    current_shell = pwd.getpwnam(getpass.getuser()).pw_shell.split("/")[-1]
    source_designator = "source" if current_shell in ["bash", "sh"] else "source_zsh"
    default_environment = ".bashrc" if current_shell in ["bash", "sh"] else ".zshrc"
    with open(config_file.absolute(), "w") as config_fd:
        config_fd.write('eval "$(_ESQUE_COMPLETE=' + source_designator + ' esque)"')
    click.echo("Autocompletion script generated to " + green_bold(str(config_file.absolute())))
    click.echo(
        "To use the autocompletion feature, simply source the contents of the script into your environment, e.g."
    )
    click.echo(
        '\t\techo -e "\\nsource '
        + str(config_file.absolute())
        + '" >> '
        + str(pwd.getpwnam(getpass.getuser()).pw_dir)
        + "/"
        + default_environment
    )


@config.command("edit", short_help="Edit esque config file.")
@default_options
def config_edit(state: State):
    """Opens the user's esque config file in the default editor."""
    old_yaml = config_path().read_text()
    new_yaml, _ = edit_yaml(old_yaml, validator=validation.validate_esque_config)
    config_path().write_text(new_yaml)


@config.command("migrate")
@default_options
def config_migrate(state: State):
    """Migrate esque config to current version.

    If the user's esque config file is from a previous version, migrate it to the current version.
    A backup of the original config file is created with the same name and the extension .bak"""
    new_path, backup = migration.migrate(config_path())
    click.echo(f"Your config has been migrated and is now at {new_path}. A backup has been created at {backup}.")


@create.command("topic")
@click.argument("topic-name", metavar="TOPIC_NAME", callback=fallback_to_stdin, required=False)
@click.option(
    "-l",
    "--like",
    metavar="<template_topic>",
    help="Topic to use as template.",
    autocompletion=list_topics,
    required=False,
)
@default_options
def create_topic(state: State, topic_name: str, like: str):
    """Create a topic.

    Create a topic called TOPIC_NAME with the option of providing a template topic, <template_topic>,
    from which all the configuration options will be copied.
    """
    if not ensure_approval("Are you sure?", no_verify=state.no_verify):
        click.echo("Aborted!")
        return

    topic_controller = state.cluster.topic_controller
    if like:
        template_config = topic_controller.get_cluster_topic(like)
        topic = Topic(
            topic_name, template_config.num_partitions, template_config.replication_factor, template_config.config
        )
    else:
        topic = Topic(topic_name)
    topic_controller.create_topics([topic])
    click.echo(click.style(f"Topic with name '{topic.name}' successfully created.", fg="green"))


@edit.command("topic")
@click.argument("topic-name", required=True, autocompletion=list_topics)
@default_options
def edit_topic(state: State, topic_name: str):
    """Edit a topic.

    Open the topic's configuration in the default editor. If the user saves upon exiting the editor,
    all the given changes will be applied to the topic.
    """
    controller = state.cluster.topic_controller
    topic = state.cluster.topic_controller.get_cluster_topic(topic_name)

    _, new_conf = edit_yaml(topic.to_yaml(only_editable=True), validator=validation.validate_editable_topic_config)

    local_topic = copy_to_local(topic)
    local_topic.update_from_dict(new_conf)
    diff = controller.diff_with_cluster(local_topic)
    if not diff.has_changes:
        click.echo("Nothing changed.")
        return

    click.echo(pretty_topic_diffs({topic_name: diff}))
    if ensure_approval("Are you sure?"):
        controller.alter_configs([local_topic])
    else:
        click.echo("Canceled!")


@set_.command("offsets")
@click.argument("consumer-id", callback=fallback_to_stdin, type=click.STRING, required=True)
@click.option(
    "-t",
    "--topic-name",
    help="Regular expression describing the topic name (default: all subscribed topics)",
    type=click.STRING,
    required=False,
)
@click.option("--offset-to-value", help="Set offset to the specified value", type=click.INT, required=False)
@click.option("--offset-by-delta", help="Shift offset by specified value", type=click.INT, required=False)
@click.option(
    "--offset-to-timestamp",
    help="Set offset to that of the first message with timestamp on or after the specified timestamp in format "
    "YYYY-MM-DDTHH:mm:ss, i.e. skip all messages before this timestamp."
    f" {red_bold('Beware! This can be a really expensive operation.')}",
    type=click.STRING,
    required=False,
)
@click.option(
    "--offset-from-group", help="Copy all offsets from an existing consumer group.", type=click.STRING, required=False
)
@default_options
def set_offsets(
    state: State,
    consumer_id: str,
    topic_name: str,
    offset_to_value: int,
    offset_by_delta: int,
    offset_to_timestamp: str,
    offset_from_group: str,
):
    """Set consumer group offsets.

    Change or set the offset of a consumer group for a topic, i.e. the message number the consumer group will read next.
    This can be done by specifying an explicit offset (--offset-to-value), a delta to shift the current offset forwards
    or backwards (--offset-by-delta), a timestamp in which the offset of the first message on or after the timestamp is
    taken (--offset-by-timestamp), or a group from which to copy the offsets from. In the case that the consumer group
    reads from more than one topic, a regular expression can be given to specify the offset of which topic to change.
    NOTE: the default is to change the offset for all topics."""
    logger = logging.getLogger(__name__)
    consumergroup_controller = ConsumerGroupController(state.cluster)
    offset_plan = consumergroup_controller.create_consumer_group_offset_change_plan(
        consumer_id=consumer_id,
        topic_name=topic_name if topic_name else ".*",
        offset_to_value=offset_to_value,
        offset_by_delta=offset_by_delta,
        offset_to_timestamp=offset_to_timestamp,
        offset_from_group=offset_from_group,
    )

    if offset_plan and len(offset_plan) > 0:
        click.echo(green_bold("Proposed offset changes: "))
        offset_plan.sort(key=attrgetter("topic_name", "partition_id"))
        for topic_name, group in groupby(offset_plan, attrgetter("topic_name")):
            group = list(group)
            max_proposed = max(len(str(elem.proposed_offset)) for elem in group)
            max_current = max(len(str(elem.current_offset)) for elem in group)
            for plan_element in group:
                new_offset = str(plan_element.proposed_offset).rjust(max_proposed)
                format_args = dict(
                    topic_name=plan_element.topic_name,
                    partition_id=plan_element.partition_id,
                    current_offset=plan_element.current_offset,
                    new_offset=new_offset if plan_element.offset_equal else red_bold(new_offset),
                    max_current=max_current,
                )
                click.echo(
                    "Topic: {topic_name}, partition {partition_id:2}, current offset: {current_offset:{max_current}}, new offset: {new_offset}".format(
                        **format_args
                    )
                )
        if ensure_approval("Are you sure?", no_verify=state.no_verify):
            consumergroup_controller.edit_consumer_group_offsets(consumer_id=consumer_id, offset_plan=offset_plan)
    else:
        logger.info("No changes proposed.")
        return


@delete.command("topic")
@click.argument(
    "topic-name",
    metavar="TOPIC_NAME",
    callback=fallback_to_stdin,
    required=False,
    type=click.STRING,
    autocompletion=list_topics,
)
@default_options
def delete_topic(state: State, topic_name: str):
    """Delete a topic

    WARNING: This command cannot be undone, and all data in the topic will be lost.
    """
    topic_controller = state.cluster.topic_controller
    if ensure_approval("Are you sure?", no_verify=state.no_verify):
        topic_controller.delete_topic(Topic(topic_name))

        assert topic_name not in (t.name for t in topic_controller.list_topics(get_topic_objects=False))

    click.echo(click.style(f"Topic with name '{topic_name}' successfully deleted.", fg="green"))


@esque.command("apply")
@click.option("-f", "--file", metavar="<file>", help="Config file path.", required=True)
@default_options
def apply(state: State, file: str):
    """Apply a set of topic configurations.

    Create new topics and apply changes to existing topics, as specified in the config yaml file <file>.
    """

    # Get topic data based on the YAML
    yaml_topic_configs = yaml.safe_load(open(file)).get("topics")
    yaml_topics = [Topic.from_dict(conf) for conf in yaml_topic_configs]
    yaml_topic_names = [t.name for t in yaml_topics]
    if not len(yaml_topic_names) == len(set(yaml_topic_names)):
        raise ValidationException("Duplicate topic names in the YAML!")

    # Get topic data based on the cluster state
    topic_controller = state.cluster.topic_controller
    cluster_topics = topic_controller.list_topics(search_string="|".join(yaml_topic_names))
    cluster_topic_names = [t.name for t in cluster_topics]

    # Calculate changes
    to_create = [yaml_topic for yaml_topic in yaml_topics if yaml_topic.name not in cluster_topic_names]
    to_edit = [
        yaml_topic
        for yaml_topic in yaml_topics
        if yaml_topic not in to_create and topic_controller.diff_with_cluster(yaml_topic).has_changes
    ]
    to_edit_diffs = {t.name: topic_controller.diff_with_cluster(t) for t in to_edit}
    to_ignore = [yaml_topic for yaml_topic in yaml_topics if yaml_topic not in to_create and yaml_topic not in to_edit]

    # Sanity check - the 3 groups of topics should be complete and have no overlap
    assert (
        set(to_create).isdisjoint(set(to_edit))
        and set(to_create).isdisjoint(set(to_ignore))
        and set(to_edit).isdisjoint(set(to_ignore))
        and len(to_create) + len(to_edit) + len(to_ignore) == len(yaml_topics)
    )

    # Print diffs so the user can check
    click.echo(pretty_unchanged_topic_configs(to_ignore))
    click.echo(pretty_new_topic_configs(to_create))
    click.echo(pretty_topic_diffs(to_edit_diffs))

    # Check for actionable changes
    if len(to_edit) + len(to_create) == 0:
        click.echo("No changes detected, aborting!")
        return

    # Warn users & abort when replication & num_partition changes are attempted
    if any(not diff.is_valid for _, diff in to_edit_diffs.items()):
        click.echo(
            "Changes to `replication_factor` and `num_partitions` can not be applied on already existing topics."
        )
        click.echo("Cancelling due to invalid changes")
        return

    # Get approval
    if not ensure_approval("Apply changes?", no_verify=state.no_verify):
        click.echo("Cancelling changes")
        return

    # apply changes
    topic_controller.create_topics(to_create)
    topic_controller.alter_configs(to_edit)

    # output confirmation
    changes = {"unchanged": len(to_ignore), "created": len(to_create), "changed": len(to_edit)}
    click.echo(click.style(pretty({"Successfully applied changes": changes}), fg="green"))


@describe.command("topic")
@click.argument(
    "topic-name",
    metavar="TOPIC_NAME",
    callback=fallback_to_stdin,
    required=False,
    type=click.STRING,
    autocompletion=list_topics,
)
@click.option(
    "--consumers",
    "-c",
    required=False,
    is_flag=True,
    default=False,
    help="Will output the consumer groups reading from this topic."
    f" {red_bold('Beware! This can be a really expensive operation.')}",
)
@output_format_option
@default_options
def describe_topic(state: State, topic_name: str, consumers: bool, output_format: str):
    """Describe a topic.

    Returns information on a given topic and its partitions, with the option of including
    all consumer groups that read from the topic.
    """
    topic = state.cluster.topic_controller.get_cluster_topic(topic_name)

    output_dict = {
        "topic": topic_name,
        "partitions": [partition.as_dict() for partition in topic.partitions],
        "config": topic.config,
    }

    if consumers:
        consumergroup_controller = ConsumerGroupController(state.cluster)
        groups = consumergroup_controller.list_consumer_groups()

        consumergroups = [
            group_name
            for group_name in groups
            if topic_name in consumergroup_controller.get_consumergroup(group_name).topics
        ]

        output_dict["consumergroups"] = consumergroups
    click.echo(format_output(output_dict, output_format))


@get.command("watermarks", short_help="Return watermarks by topic.")
@click.option(
    "-t", "--topic-name", metavar="<topic_name>", required=False, type=click.STRING, autocompletion=list_topics
)
@output_format_option
@default_options
def get_watermarks(state: State, topic_name: str, output_format: str):
    """Returns the high and low watermarks for <topic_name>, or if not specified, all topics."""
    # TODO: Gathering of all watermarks takes super long
    topics = state.cluster.topic_controller.list_topics(search_string=topic_name)

    watermarks = {topic.name: max(v for v in topic.watermarks.values()) for topic in topics}

    click.echo(format_output(watermarks, output_format))


@describe.command("broker", short_help="Describe a broker.")
@click.argument("broker", metavar="BROKER", callback=fallback_to_stdin, autocompletion=list_brokers, required=False)
@output_format_option
@default_options
def describe_broker(state: State, broker: str, output_format: str):
    """Return configuration options for broker BROKER. BROKER can be given with broker id (integer),
    the host name (if hostname is unique), or socket address ('hostname:port')"""
    if broker.isdigit():
        broker = Broker.from_id(state.cluster, broker).describe()
    elif ":" not in broker:
        broker = Broker.from_host(state.cluster, broker).describe()
    else:
        try:
            host, port = broker.split(":")
            broker = Broker.from_host_and_port(state.cluster, host, int(port)).describe()
        except ValueError:
            raise ValidationException("BROKER must either be the broker id, the hostname, or in the form 'host:port'")

    click.echo(format_output(broker, output_format))


@describe.command("consumergroup", short_help="Describe a consumer group.")
@click.argument("consumer-id", callback=fallback_to_stdin, autocompletion=list_consumergroups, required=True)
@click.option(
    "--all-partitions",
    help="List status for all topic partitions instead of just summarizing each topic.",
    default=False,
    is_flag=True,
)
@output_format_option
@default_options
def describe_consumergroup(state: State, consumer_id: str, all_partitions: bool, output_format: str):
    """Return information on group coordinator, offsets, watermarks, lag, and various metadata
     for consumer group CONSUMER_GROUP."""
    consumer_group = ConsumerGroupController(state.cluster).get_consumergroup(consumer_id)
    consumer_group_desc = consumer_group.describe(verbose=all_partitions)

    click.echo(format_output(consumer_group_desc, output_format))


@get.command("brokers")
@output_format_option
@default_options
def get_brokers(state: State, output_format: str):
    """List all brokers.

    Return the broker id's and socket addresses of all the brokers in the kafka cluster defined in the current context.
    """
    brokers = Broker.get_all(state.cluster)
    broker_ids_and_hosts = [f"{broker.broker_id}: {broker.host}:{broker.port}" for broker in brokers]
    click.echo(format_output(broker_ids_and_hosts, output_format))


@get.command("consumergroups")
@output_format_option
@default_options
def get_consumergroups(state: State, output_format: str):
    """List all consumer groups."""
    groups = ConsumerGroupController(state.cluster).list_consumer_groups()
    click.echo(format_output(groups, output_format))


@get.command("topics")
@click.option("-p", "--prefix", type=click.STRING, autocompletion=list_topics)
@output_format_option
@default_options
def get_topics(state: State, prefix: str, output_format: str):
    """List all topics."""
    topics = state.cluster.topic_controller.list_topics(search_string=prefix, get_topic_objects=False)
    topic_names = [topic.name for topic in topics]
    click.echo(format_output(topic_names, output_format))


@esque.command("consume")
@click.argument("topic", autocompletion=list_topics)
@click.option(
    "-d", "--directory", metavar="<directory>", help="Sets the directory to write the messages to.", type=click.STRING
)
@click.option(
    "-f",
    "--from",
    "from_context",
    metavar="<source_ctx>",
    help="Source context. If not provided, the current context will be used.",
    autocompletion=list_contexts,
    type=click.STRING,
    required=False,
)
@click.option(
    "-n", "--numbers", metavar="<n>", help="Number of messages.", type=click.INT, default=sys.maxsize, required=False
)
@click.option(
    "-m",
    "--match",
    metavar="<filter_expression>",
    help="Message filtering expression.",
    type=click.STRING,
    required=False,
)
@click.option("--last/--first", help="Start consuming from the earliest or latest offset in the topic.", default=False)
@click.option("-a", "--avro", help="Set this flag if the topic contains avro data.", default=False, is_flag=True)
@click.option(
    "-c",
    "--consumergroup",
    metavar="<consumer_group>",
    help="Consumer group to store the offset in.",
    type=click.STRING,
    autocompletion=list_consumergroups,
    default=None,
    required=False,
)
@click.option(
    "--preserve-order",
    help="Preserve the order of messages, regardless of their partition.",
    default=False,
    is_flag=True,
)
@click.option("--stdout", "write_to_stdout", help="Write messages to STDOUT.", default=False, is_flag=True)
@default_options
def consume(
    state: State,
    topic: str,
    from_context: str,
    numbers: int,
    match: str,
    last: bool,
    avro: bool,
    directory: str,
    consumergroup: str,
    preserve_order: bool,
    write_to_stdout: bool,
):
    """Consume messages from a topic.

    Read messages from a given topic in a given context. These messages can either be written
    to files in an automatically generated directory (default behavior), or to STDOUT.

    \b
    EXAMPLES:
    # Consume the first 10 messages from TOPIC in the current context and print them to STDOUT in order.
    esque consume --first -n 10 --preserve-order --stdout TOPIC

    \b
    # Consume <n> messages, starting from the 10th, from TOPIC in the <source_ctx> context and write them to files.
    esque consume --match "message.offset > 9" -n <n> TOPIC -f <source_ctx>

    \b
    # Copy source_topic in first context to destination_topic in second-context.
    esque consume -f first-context --stdout source_topic | esque produce -t second-context --stdin destination_topic
    """
    current_timestamp_milliseconds = int(round(time.time() * 1000))
    consumergroup_prefix = "group_for_"

    if directory and write_to_stdout:
        raise ValueError("Cannot write to a directory and STDOUT, please pick one!")

    if not from_context:
        from_context = state.config.current_context
    state.config.context_switch(from_context)

    if topic not in map(attrgetter("name"), state.cluster.topic_controller.list_topics(get_topic_objects=False)):
        raise TopicDoesNotExistException(f"Topic {topic} does not exist!", -1)

    if not consumergroup:
        consumergroup = consumergroup_prefix + topic + "_" + str(current_timestamp_milliseconds)
    if not directory:
        directory = Path() / "messages" / topic / str(current_timestamp_milliseconds)
    output_directory = Path(directory)

    if not write_to_stdout:
        click.echo(f"Creating directory {blue_bold(str(output_directory))} if it does not exist.")
        output_directory.mkdir(parents=True, exist_ok=True)
        click.echo(f"Start consuming from topic {blue_bold(topic)} in source context {blue_bold(from_context)}.")
    if preserve_order:
        partitions = []
        for partition in state.cluster.topic_controller.get_cluster_topic(topic).partitions:
            partitions.append(partition.partition_id)
        total_number_of_consumed_messages = consume_to_file_ordered(
            output_directory=output_directory,
            topic=topic,
            group_id=consumergroup,
            partitions=partitions,
            numbers=numbers,
            avro=avro,
            match=match,
            last=last,
            write_to_stdout=write_to_stdout,
        )
    else:
        total_number_of_consumed_messages = consume_to_files(
            output_directory=output_directory,
            topic=topic,
            group_id=consumergroup,
            numbers=numbers,
            avro=avro,
            match=match,
            last=last,
            write_to_stdout=write_to_stdout,
        )

    if not write_to_stdout:
        click.echo(f"Output generated to {blue_bold(str(output_directory))}")
        if total_number_of_consumed_messages == numbers or numbers == sys.maxsize:
            click.echo(blue_bold(str(total_number_of_consumed_messages)) + " messages consumed.")
        else:
            click.echo(
                "Only found "
                + bold(str(total_number_of_consumed_messages))
                + " messages in topic, out of "
                + blue_bold(str(numbers))
                + " required."
            )


@esque.command("produce")
@click.argument("topic", autocompletion=list_topics)
@click.option(
    "-d",
    "--directory",
    metavar="<directory>",
    help="Directory containing Kafka messages.",
    type=click.STRING,
    required=False,
)
@click.option(
    "-t",
    "--to",
    "to_context",
    metavar="<destination_ctx>",
    help="Destination context.",
    type=click.STRING,
    autocompletion=list_contexts,
    required=False,
)
@click.option(
    "-m",
    "--match",
    metavar="<filter_expresion>",
    help="Message filtering expression.",
    type=click.STRING,
    required=False,
)
@click.option("-a", "--avro", help="Set this flag if the topic contains avro data.", default=False, is_flag=True)
@click.option(
    "--stdin", "read_from_stdin", help="Read messages from STDIN instead of a directory.", default=False, is_flag=True
)
@click.option(
    "-y",
    "--ignore-errors",
    "ignore_stdin_errors",
    help="Only when reading from STDIN. If JSON validation fails, write the malformed JSON as a string in message value"
    " (without key and specified partition assignment).",
    default=False,
    is_flag=True,
)
@default_options
def produce(
    state: State,
    topic: str,
    to_context: str,
    directory: str,
    avro: bool,
    match: str = None,
    read_from_stdin: bool = False,
    ignore_stdin_errors: bool = False,
):
    """Produce messages to a topic.

       Write messages to a given topic in a given context. These messages can come from either a directory <directory>
       containing files corresponding to the different partitions or from STDIN.

       \b
       EXAMPLES:
       # Write all messages from the files in <directory> to TOPIC in the <destination_ctx> context.
       esque produce -d <directory> -t <destination_ctx> TOPIC

       \b
       # Start environment in terminal to write messages to TOPIC in the <destination_ctx> context.
       esque produce --stdin -f <destination_ctx> -y TOPIC

       \b
       # Copy source_topic to destination_topic.
       esque consume -f first-context --stdout source_topic | esque produce -t second-context --stdin destination_topic
       """
    if directory is None and not read_from_stdin:
        raise ValueError("You have to provide a directory or use the --stdin flag.")

    if directory is not None:
        input_directory = Path(directory)
        if not input_directory.exists():
            raise ValueError(f"Directory {directory} does not exist!")

    if not to_context:
        to_context = state.config.current_context
    state.config.context_switch(to_context)

    topic_controller = state.cluster.topic_controller
    if topic not in map(attrgetter("name"), topic_controller.list_topics(get_topic_objects=False)):
        click.echo(f"Topic {blue_bold(topic)} does not exist in context {blue_bold(to_context)}.")
        if ensure_approval(f"Would you like to create it now?"):
            topic_controller.create_topics([Topic(topic)])
        else:
            raise TopicDoesNotExistException(f"Topic {topic} does not exist!", -1)

    stdin = click.get_text_stream("stdin")
    if read_from_stdin and isatty(stdin):
        click.echo(
            "Type the messages to produce, "
            + ("in JSON format, " if not ignore_stdin_errors else "")
            + blue_bold("one per line")
            + ". End with "
            + blue_bold("CTRL+D")
            + "."
        )
    elif read_from_stdin and not isatty(stdin):
        click.echo(f"Reading messages from an external source, {blue_bold('one per line')}).")
    else:
        click.echo(
            f"Producing from directory {blue_bold(str(directory))} to topic {blue_bold(topic)}"
            f" in target context {blue_bold(to_context)}"
        )
    producer = ProducerFactory().create_producer(
        topic_name=topic,
        input_directory=input_directory if not read_from_stdin else None,
        avro=avro,
        match=match,
        ignore_stdin_errors=ignore_stdin_errors,
    )
    total_number_of_messages_produced = producer.produce()
    click.echo(
        green_bold(str(total_number_of_messages_produced))
        + " messages successfully produced to topic "
        + blue_bold(topic)
        + " in context "
        + blue_bold(to_context)
        + "."
    )


@esque.command("ping")
@click.option("-t", "--times", help="Number of pings.", default=10)
@click.option("-w", "--wait", help="Seconds to wait between pings.", default=1)
@default_options
def ping(state: State, times: int, wait: int):
    """Test the connection to the kafka cluster.

    Ping the kafka cluster by writing messages to and reading messages from it.
    After the specified number of "pings", return the minimum, maximum, and average time for the round trip.
    """
    topic_controller = state.cluster.topic_controller
    deltas = []
    try:
        try:
            topic_controller.create_topics([Topic(PING_TOPIC)])
        except TopicAlreadyExistsException:
            pass
        producer = PingProducer(PING_TOPIC)
        consumer = ConsumerFactory().create_ping_consumer(group_id=PING_GROUP_ID, topic_name=PING_TOPIC)
        click.echo(f"Pinging with {state.cluster.bootstrap_servers}.")

        for i in range(times):
            producer.produce()
            _, delta = consumer.consume()
            deltas.append(delta)
            click.echo(f"m_seq={i} time={delta:.2f}ms")
            sleep(wait)
    except KeyboardInterrupt:
        return
    topic_controller.delete_topic(Topic(PING_TOPIC))
    click.echo("--- statistics ---")
    click.echo(f"{len(deltas)} messages sent/received.")
    click.echo(f"min/avg/max = {min(deltas):.2f}/{(sum(deltas) / len(deltas)):.2f}/{max(deltas):.2f} ms")



import re
from abc import ABC

import esque.ruleparser.exception as ex
import esque.ruleparser.helpers as h
from esque.ruleparser.exception import ExpressionSyntaxError, MalformedExpressionError, RuleTreeInvalidError
from esque.ruleparser.expressionelement import (
    ArithmeticBinaryOperator,
    ComparisonBinaryOperator,
    ComparisonUnaryOperator,
    Field,
    GenericOperator,
    Literal,
    Operator,
    OperatorType,
    ParenthesisOperator,
)
from esque.ruleparser.fieldeval import FieldEval


class RuleTree:
    __postfix_expression = []
    __infix_expression = []
    __tree_valid = False
    __expression_string = ""

    def __init__(self, expression_string: str):
        self.__infix_expression = []
        self.__postfix_expression = []
        self.__tree_valid = False
        self.__expression_string = expression_string
        parser = ExpressionParser(self.__expression_string)
        self.__infix_expression = parser.parse_expression()
        self.__convert_infix_to_postfix()

    def __convert_infix_to_postfix(self):
        """
        Responsible for converting the infix expression notation (i.e. a+b-c) to postfix expression notation (i.e. ab+c-).
        :return: No return value.
        """
        operator_stack = []
        parenthesis_balance_counter = 0
        rank = 0
        element: ABC
        for element in self.__infix_expression:
            if isinstance(element, ParenthesisOperator):
                operator: Operator = element
                if operator.literal == "(":
                    operator_stack.append(operator)
                    parenthesis_balance_counter = parenthesis_balance_counter + 1
                else:
                    try:
                        while operator_stack[-1].literal != "(":
                            self.__postfix_expression.append(operator_stack.pop())
                        operator_stack.pop()
                        parenthesis_balance_counter = parenthesis_balance_counter - 1
                    except IndexError:
                        raise ExpressionSyntaxError("Malformed parentheses in pattern")
            elif isinstance(element, Operator):
                operator: Operator = element
                while (
                    len(operator_stack) > 0
                    and operator_stack[-1].generic_operator.stack_priority >= operator.generic_operator.input_priority
                ):
                    self.__postfix_expression.append(operator_stack.pop())
                    rank = rank - 1
                if rank < 0:
                    raise ExpressionSyntaxError("Malformed pattern at element " + operator.literal)
                operator_stack.append(operator)
            else:
                self.__postfix_expression.append(element)
                rank = rank + 1
        while len(operator_stack) > 0:
            if operator_stack[-1].generic_operator.type != OperatorType.PARENTHESIS:
                self.__postfix_expression.append(operator_stack.pop())
            elif operator_stack[-1].literal == "(":
                parenthesis_balance_counter = parenthesis_balance_counter + 1
            elif operator_stack[-1].literal == ")":
                parenthesis_balance_counter = parenthesis_balance_counter - 1
        if parenthesis_balance_counter != 0:
            raise ExpressionSyntaxError("Malformed parentheses in pattern")
        self.__tree_valid = True

    def evaluate(self, message):
        """
        Evaluates the parsed expression against a Kafka message (if required).
        :param message: The Kafka message used for evaluation. If the evaluated expression contains no references to the message, this parameter can be set to None.
        :return:
        The final value. If any problems are encountered, this method raises a MalformedExpressionError.
        """
        operand_stack = []
        field_evaluator = FieldEval(message)
        if not self.__tree_valid:
            raise RuleTreeInvalidError(
                "The rule tree is not valid. The expression must be parsed before evaluating the tree."
            )
        for tree_element in self.__postfix_expression:
            if isinstance(tree_element, ArithmeticBinaryOperator) or isinstance(
                tree_element, ComparisonBinaryOperator
            ):
                # we need to have 2 operands on stack at this point, otherwise the expression is malformed
                if len(operand_stack) < 2:
                    raise MalformedExpressionError("Operator " + tree_element.literal + " requires 2 operands.")
                evaluation_result = tree_element.evaluate(str(operand_stack[-2]), str(operand_stack[-1]))
                operand_stack.pop()
                operand_stack.pop()
                operand_stack.append(evaluation_result)
            elif isinstance(tree_element, ComparisonUnaryOperator):
                # we need to have 1 operand on stack at this point, otherwise the expression is malformed
                if len(operand_stack) < 1:
                    raise MalformedExpressionError("Operator " + tree_element.literal + " requires 1 operand.")
                evaluation_result = tree_element.evaluate(str(operand_stack[-1]))
                operand_stack.pop()
                operand_stack.append(evaluation_result)
            elif isinstance(tree_element, Literal):
                operand_stack.append(tree_element.value)
            elif isinstance(tree_element, Field):
                operand_stack.append(field_evaluator.evaluate_field(tree_element.field_name))
        # in the end, if we have only one element on the stack, that is our result
        # otherwise, the expression is malformed
        if len(operand_stack) == 1:
            return operand_stack[0]
        else:
            raise MalformedExpressionError("The expression is malformed.")


class ExpressionParser:

    RE_SPACE: str = h.zero_or_more('[" "]|\t')
    RE_LITERALS = [
        "[1-9][0-9]{3}(-([0-9]{2})){2}T([0-9]{2})(:([0-9]{2})){2}",  # DATE TIME STAMP: YYYY-MM-DDTHH:mm:ss
        "[1-9][0-9]{3}(-([0-9]{2})){2}",  # DATESTAMP: YYYY-MM-DD
        "[-+]?[1-9]+[0-9]*",  # INTEGER
        "[+-]?[0-9]+.[0-9]*[eE]*[+-]?[0-9]*",  # FLOATING POINT
        "(true|false)",  # BOOLEAN
        "[a-zA-Z0-9\\-\\._\\%]+",  # STRING
    ]

    expression: str = ""
    arithmetic_operators_pattern = None
    comparison_operators_pattern = None
    parenthesis_pattern = None
    fields_pattern = None
    literals_pattern = None

    def __init__(self, expression: str = ""):
        self.expression = expression.strip()
        self.arithmetic_operators_pattern = re.compile(
            ExpressionParser.RE_SPACE
            + h.either_of(
                list(
                    map(
                        lambda x: x.regex,
                        filter(lambda y: y.type == OperatorType.ARITHMETIC, Operator.OPERATORS.values()),
                    )
                )
            ),
            flags=re.IGNORECASE | re.UNICODE,
        )
        self.comparison_operators_pattern = re.compile(
            ExpressionParser.RE_SPACE
            + h.either_of(
                list(
                    map(
                        lambda x: x.regex,
                        filter(lambda y: y.type == OperatorType.COMPARISON, Operator.OPERATORS.values()),
                    )
                )
            ),
            flags=re.IGNORECASE | re.UNICODE,
        )
        self.parenthesis_pattern = re.compile(
            ExpressionParser.RE_SPACE
            + h.either_of(
                list(
                    map(
                        lambda x: x.regex,
                        filter(lambda y: y.type == OperatorType.PARENTHESIS, Operator.OPERATORS.values()),
                    )
                )
            ),
            flags=re.IGNORECASE | re.UNICODE,
        )
        self.fields_pattern = re.compile(
            ExpressionParser.RE_SPACE + h.either_of(Operator.FIELDS.values()), flags=re.IGNORECASE | re.UNICODE
        )
        self.literals_pattern = re.compile(
            ExpressionParser.RE_SPACE + h.either_of(ExpressionParser.RE_LITERALS), flags=re.IGNORECASE
        )

    def parse_expression(self):
        """
        Parses the input expression and returns a list of discovered elements.
        :return: The list of discovered (parsed elements). The list is populated in the order the elements are found (i.e. "infix" notation, a+b+c+d)
        """
        parsed_elements = []
        if self.expression.strip() == "":
            raise ex.EmptyExpressionError("The expression string contains no expression.")
        previous_position = 0
        match_end = 0

        while previous_position < len(self.expression):
            match_arithmetic_operator = self.arithmetic_operators_pattern.match(
                string=self.expression, pos=previous_position
            )
            match_comparison_operator = self.comparison_operators_pattern.match(
                string=self.expression, pos=previous_position
            )
            match_parenthesis_operator = self.parenthesis_pattern.match(string=self.expression, pos=previous_position)
            match_field = self.fields_pattern.match(string=self.expression, pos=previous_position)
            match_literal = self.literals_pattern.match(string=self.expression, pos=previous_position)
            match_found = False
            if match_arithmetic_operator is not None:
                if match_arithmetic_operator.start(0) == previous_position:
                    match_found = True
                    match_end = match_arithmetic_operator.end(0)
                    content = match_arithmetic_operator.group(0).strip()
                    generic_operator: GenericOperator = list(
                        filter(lambda x: x.regex.replace("\\", "") == content, Operator.OPERATORS.values())
                    )[0]
                    parsed_elements.append(
                        ArithmeticBinaryOperator(literal=content, generic_operator=generic_operator)
                    )
            if not match_found and match_comparison_operator is not None:
                if match_comparison_operator.start(0) == previous_position:
                    match_found = True
                    match_end = match_comparison_operator.end(0)
                    content = match_comparison_operator.group(0).strip()
                    generic_operator: GenericOperator = list(
                        filter(lambda x: x.regex.replace("\\", "") == content, Operator.OPERATORS.values())
                    )[0]
                    if generic_operator.is_unary:
                        parsed_elements.append(
                            ComparisonUnaryOperator(literal=content, generic_operator=generic_operator)
                        )
                    else:
                        parsed_elements.append(
                            ComparisonBinaryOperator(literal=content, generic_operator=generic_operator)
                        )
            if not match_found and match_parenthesis_operator is not None:
                if match_parenthesis_operator.start(0) == previous_position:
                    match_found = True
                    match_end = match_parenthesis_operator.end(0)
                    content = match_parenthesis_operator.group(0).strip()
                    generic_operator: GenericOperator = list(
                        filter(lambda x: x.regex.replace("\\", "") == content, Operator.OPERATORS.values())
                    )[0]
                    parsed_elements.append(ParenthesisOperator(literal=content, generic_operator=generic_operator))
            if not match_found and match_field is not None:
                if match_field.start(0) == previous_position:
                    match_found = True
                    match_end = match_field.end(0)
                    content = match_field.group(0).strip()
                    parsed_elements.append(Field(field_name=content))
            if not match_found and match_literal is not None:
                if match_literal.start(0) == previous_position:
                    match_found = True
                    match_end = match_literal.end(0)
                    content = match_literal.group(0).strip()
                    parsed_elements.append(Literal(value=content))
            if not match_found:
                raise ex.ExpressionSyntaxError("Unrecognized symbols at position: " + str(previous_position))
            else:
                previous_position = match_end
        return parsed_elements



from pathlib import Path

import toml

pyproject = Path(__file__).parent.parent / "pyproject.toml"
assert pyproject.exists(), "pyproject.toml doesn't exist"
data = toml.loads(pyproject.read_text())
__version__ = data["tool"]["poetry"]["version"]



import itertools as it
import json
import pathlib
import pickle
import struct
from io import BytesIO
from typing import Any, Dict, Iterable, List, NamedTuple, Optional, Tuple

import click
import fastavro
from avro.schema import RecordSchema
from confluent_kafka.avro import loads as load_schema
from confluent_kafka.cimpl import Message

from esque.clients.schemaregistry import SchemaRegistryClient
from esque.messages.message import FileReader, FileWriter, KafkaMessage, MessageHeader, StdOutWriter

KEY_SCHEMA_FOLDER_NAME = "key_schema"
VALUE_SCHEMA_FOLDER_NAME = "value_schema"


class DecodedAvroMessage(NamedTuple):
    key: Any
    value: Any
    partition: int
    key_schema_id: int
    value_schema_id: int
    headers: List[MessageHeader] = []


class AvroFileWriter(FileWriter):
    def __init__(
        self, schema_registry_client: SchemaRegistryClient, directory: pathlib.Path, partition: Optional[str] = None
    ):
        super().__init__(directory, partition)
        self.directory = directory
        self.schema_registry_client = schema_registry_client
        self.current_key_schema_id = None
        self.current_value_schema_id = None
        self.schema_dir_name = None
        self.schema_version = it.count(1)
        self.open_mode = "wb+"
        self.avro_decoder = AvroMessageDecoder(self.schema_registry_client)

    def write_message(self, message: Message):
        key_schema_id, value_schema_id, decoded_message, serializable_message = self.avro_decoder.decode_message_from_avro(
            message
        )

        self._dump_schema(value_schema_id, VALUE_SCHEMA_FOLDER_NAME)
        self._dump_schema(key_schema_id, KEY_SCHEMA_FOLDER_NAME)
        pickle.dump(serializable_message, self.file)

    def _dump_schema(self, schema_id: int, schema_folder_name: str):
        if schema_id == -1:
            return
        schema_file_path = get_schema_file_path(self.directory, schema_id, schema_folder_name)
        folder_path = schema_file_path.parent
        if not folder_path.exists():
            folder_path.mkdir(exist_ok=True)
        schema_file_path.write_text(
            json.dumps(self.schema_registry_client.get_schema_from_id(schema_id).original_schema), encoding="utf-8"
        )


class StdOutAvroWriter(StdOutWriter):
    def __init__(self, schema_registry_client):
        self.avro_decoder = AvroMessageDecoder(schema_registry_client=schema_registry_client)

    def write_message(self, message: Message):
        _, _, _, plaintext_message = self.avro_decoder.decode_message_from_avro(message)
        click.echo(json.dumps(plaintext_message))


class AvroFileReader(FileReader):
    def __init__(self, directory: pathlib.Path, partition: Optional[str] = None):
        super().__init__(directory, partition)
        self.open_mode = "rb"

    def read_message_from_file(self) -> Iterable[KafkaMessage]:
        while True:
            try:
                record = pickle.load(self.file)
            except EOFError:
                return

            key_schema, value_schema = get_schemata(self.directory, record["key_schema_id"], record["value_schema_id"])
            yield KafkaMessage(
                json.dumps(record["key"]) if record["key"] is not None else None,
                json.dumps(record["value"]) if record["value"] is not None else None,
                record["partition"],
                key_schema,
                value_schema,
            )


class AvroMessageDecoder:
    def __init__(self, schema_registry_client: SchemaRegistryClient):
        self.schema_registry_client = schema_registry_client

    def decode_bytes(self, raw_data: Optional[bytes]) -> Tuple[int, Optional[Dict]]:
        if raw_data is None:
            return -1, None

        with BytesIO(raw_data) as fake_stream:
            schema_id = extract_schema_id(fake_stream.read(5))
            parsed_schema = self.schema_registry_client.get_schema_from_id(schema_id).parsed_schema
            record = fastavro.schemaless_reader(fake_stream, parsed_schema)
        return schema_id, record

    def decode_message_from_avro(self, message: Message) -> Tuple[int, int, DecodedAvroMessage, Dict]:
        key_schema_id, decoded_key = self.decode_bytes(message.key())
        value_schema_id, decoded_value = self.decode_bytes(message.value())

        headers = []
        if message.headers():
            for header_key, header_value in message.headers():
                headers.append(
                    MessageHeader(key=header_key, value=header_value.decode("utf-8") if header_value else None)
                )
        decoded_message = DecodedAvroMessage(
            decoded_key, decoded_value, message.partition(), key_schema_id, value_schema_id, headers=headers
        )

        serializable_message = {
            "key": decoded_message.key,
            "value": decoded_message.value,
            "partition": decoded_message.partition,
            "key_schema_id": key_schema_id if key_schema_id != -1 else None,
            "value_schema_id": value_schema_id if value_schema_id != -1 else None,
        }

        return key_schema_id, value_schema_id, decoded_message, serializable_message


def extract_schema_id(message: bytes) -> int:
    magic_byte, schema_id = struct.unpack(">bI", message[:5])
    assert magic_byte == 0, f"Wrong magic byte ({magic_byte}), no AVRO message."
    return schema_id


def get_schemata(
    base_directory: pathlib.Path, key_schema_id: Optional[int], value_schema_id: Optional[int]
) -> Tuple[Optional[RecordSchema], Optional[RecordSchema]]:
    return (
        (None if key_schema_id is None else get_schema(base_directory, key_schema_id, KEY_SCHEMA_FOLDER_NAME)),
        (None if value_schema_id is None else get_schema(base_directory, value_schema_id, VALUE_SCHEMA_FOLDER_NAME)),
    )


def get_schema_file_path(base_directory: pathlib.Path, schema_id: int, schema_folder_name: str) -> pathlib.Path:
    return pathlib.Path(base_directory / schema_folder_name / ("schema_id_" + str(schema_id) + ".avsc"))


def get_schema(base_directory: pathlib.Path, schema_id: int, schema_folder_name: str) -> RecordSchema:
    schema_file_path = get_schema_file_path(base_directory, schema_id, schema_folder_name)
    return load_schema(schema_file_path.read_text(encoding="utf-8"))



from collections import namedtuple
from functools import total_ordering
from typing import Any, Dict, Generator, List, Optional, Tuple, Union

import pendulum
import yaml
from pykafka.protocol.offset import OffsetPartitionResponse

from esque.resources.resource import KafkaResource

TopicDict = Dict[str, Union[int, str, Dict[str, str]]]
PartitionInfo = Dict[int, OffsetPartitionResponse]

Watermark = namedtuple("Watermark", ["high", "low"])


class Partition(KafkaResource):
    def __init__(
        self,
        partition_id: int,
        low_watermark: int,
        high_watermark: int,
        partition_isrs,
        partition_leader,
        partition_replicas: List[int],  # list of brokers holding a replica
        latest_message_timestamp: float = None,
    ):
        self.partition_id = partition_id
        self.watermark = Watermark(high_watermark, low_watermark)
        self.partition_isrs = partition_isrs
        self.partition_leader = partition_leader
        self.partition_replicas = partition_replicas
        self.latest_message_timestamp = None
        if latest_message_timestamp is not None:
            self.latest_message_timestamp = pendulum.from_timestamp(latest_message_timestamp).to_datetime_string()

    def as_dict(self):
        return {
            "partition_id": self.partition_id,
            "low_watermark": self.watermark.low,
            "high_watermark": self.watermark.high,
            "partition_isrs": self.partition_isrs,
            "partition_leader": self.partition_leader,
            "partition_replicas": self.partition_replicas,
            "latest_message_timestamp": self.latest_message_timestamp,
        }


class AttributeDiff:
    def __init__(self, remote, local):
        self.remote = remote
        self.local = local
        assert type(remote) == type(
            local
        ), f"Attributes should be given as the same type, not {type(remote)} and {type(local)}"

    @property
    def old(self):
        return self.remote

    @property
    def new(self):
        return self.local

    def __eq__(self, other: "AttributeDiff"):
        return self.remote == other.remote and self.local == other.local

    def __hash__(self):
        return hash(hash(self.remote) + hash(self.local))

    def __repr__(self):
        return f"<AttributeDiff[remote:{self.remote}, local:{self.local}]>"


class TopicDiff:
    INVALID_CHANGES = ["num_partitions", "replication_factor"]

    def __init__(self):
        self._diffs: Dict[str, AttributeDiff] = {}

    def set_diff(self, name: str, remote, local) -> "TopicDiff":

        # config values of local topics can be set as int or string
        # since all "config" variables we get from cluster topics are sent to us as strings,
        # we need to convert the local ones to string to match them correctly (exclude None's)
        local = str(local) if isinstance(remote, str) and local is not None else local

        if remote == local:
            return self

        # TODO: this should be handled correctly by checking the cluster defaults, like
        # if remote == cluster_default(name) and local is None: return
        # currently, if an attribute that was set get's un-set, it's ignored
        if local is None:
            return self

        assert type(remote) == type(
            local
        ), f"Attributes for {name} should be given as the same type, not {type(remote)} and {type(local)}"
        self._diffs[name] = AttributeDiff(remote, local)

        # allow chaining of set-calls
        return self

    @property
    def is_valid(self) -> bool:
        return set(self._diffs.get(a, None) for a in self.INVALID_CHANGES) == {None}

    @classmethod
    def from_dict(cls, diff_dict: Dict[str, AttributeDiff]) -> "TopicDiff":
        td = TopicDiff()
        td._diffs = diff_dict
        return td

    @property
    def has_changes(self) -> bool:
        return len(self._diffs.keys()) > 0

    def changes(self) -> Generator[Tuple[str, Any, Any], None, None]:
        for key, val in self._diffs.items():
            yield key, val.remote, val.local

    def __eq__(self, other: "TopicDiff") -> bool:
        return self._diffs == other._diffs

    def __repr__(self):
        return f"<TopicDiff[{str(self._diffs)}>"


def copy_to_local(topic: "Topic") -> "Topic":
    return Topic(topic.name, topic.num_partitions, topic.replication_factor, topic.config)


@total_ordering
class Topic(KafkaResource):
    def __init__(
        self,
        name: Union[str, bytes],
        num_partitions: int = None,
        replication_factor: int = None,
        config: Dict[str, str] = None,
    ):
        # Should we warn in those cases to force clients to migrate to string-only?
        if isinstance(name, bytes):
            name = name.decode("ascii")
        self.name = name

        # those settings are only used until the topic is updated from cluster
        self.__num_partitions = num_partitions
        self.__replication_factor = replication_factor

        self.config = config if config is not None else {}

        self.partition_data: Optional[List[Partition]] = None
        self.is_only_local = True

    # properties
    @property
    def partitions(self) -> List[Partition]:
        assert not self.is_only_local, "Need to update topic before updating partitions"
        return self.partition_data

    @property
    def watermarks(self) -> Dict[int, Watermark]:
        """
        Returns the low and high watermark for each partition in a topic
        """
        return {partition.partition_id: partition.watermark for partition in self.partitions}

    @property
    def num_partitions(self) -> int:
        if self.is_only_local:
            return self.__num_partitions
        return len(self.partitions)

    @property
    def replication_factor(self) -> int:
        if self.is_only_local:
            return self.__replication_factor
        partition_replication_factors = set(len(p.partition_replicas) for p in self.partitions)
        assert len(partition_replication_factors) == 1, "Different replication factors for partitions!"
        return partition_replication_factors.pop()

    # conversions and factories
    @classmethod
    def from_dict(cls, dict_object: TopicDict) -> "Topic":
        return cls(
            dict_object.get("name"),
            dict_object.get("num_partitions"),
            dict_object.get("replication_factor"),
            dict_object.get("config"),
        )

    def as_dict(self, only_editable=False) -> TopicDict:
        if only_editable:
            return {"config": self.config}
        return {
            "num_partitions": self.num_partitions,
            "replication_factor": self.replication_factor,
            "config": self.config,
        }

    def to_yaml(self, only_editable=False) -> str:
        return yaml.dump(self.as_dict(only_editable=only_editable), default_flow_style=False)

    def update_from_dict(self, new_values: TopicDict) -> None:
        for attr, value in new_values.items():
            setattr(self, attr, value)

    # object behaviour
    def __lt__(self, other: "Topic"):
        return self.name < other.name

    def __eq__(self, other: "Topic"):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)



import logging
import struct
from typing import Any, Dict, List, Optional, cast

import pykafka
from pykafka.protocol import OffsetFetchResponseV1, PartitionOffsetFetchRequest
from pykafka.protocol.admin import DescribeGroupResponse

from esque.cluster import Cluster
from esque.errors import ConsumerGroupDoesNotExistException

# TODO: Refactor this shit hole

log = logging.getLogger(__name__)


class ConsumerGroup:
    def __init__(self, id: str, cluster: Cluster):
        self.id = id
        self.cluster = cluster
        self._pykafka_group_coordinator_instance: Optional[pykafka.Broker] = None
        self.topic_controller = cluster.topic_controller

    @property
    def _pykafka_group_coordinator(self) -> pykafka.Broker:
        consumer_id = self.id.encode("UTF-8")
        if not self._pykafka_group_coordinator_instance:
            self._pykafka_group_coordinator_instance: pykafka.Broker = cast(
                pykafka.Broker, self.cluster.pykafka_client.cluster.get_group_coordinator(consumer_id)
            )

        return self._pykafka_group_coordinator_instance

    @property
    def topics(self):
        consumer_id = self.id.encode("UTF-8")

        # Get Consumers who already have an offset
        consumer_offsets = self._unpack_offset_response(
            self._pykafka_group_coordinator.fetch_consumer_group_offsets(consumer_id, preqs=[])
        )
        topic_with_offsets = set(topic.decode("UTF-8") for topic in consumer_offsets.keys())

        topics_with_members = set()
        # Get Consumers which have a member
        try:
            resp = self._pykafka_group_coordinator.describe_groups([consumer_id])
            meta = self._unpack_consumer_group_response(resp.groups[consumer_id])
            topics_with_members = set(
                member["member_metadata"]["subscription"][0].decode("UTF-8") for member in meta["members"]
            )
        except struct.error:
            pass

        return list(topic_with_offsets | topics_with_members)

    def describe(self, *, verbose=False):
        consumer_id = self.id.encode("UTF-8")
        if consumer_id not in self._pykafka_group_coordinator.list_groups().groups:
            raise ConsumerGroupDoesNotExistException(self.id)

        resp = self._pykafka_group_coordinator.describe_groups([consumer_id])
        assert len(resp.groups) == 1

        meta = self._unpack_consumer_group_response(resp.groups[consumer_id])
        consumer_offsets = self._get_consumer_offsets(consumer_id, verbose=verbose)

        return {
            "group_id": consumer_id,
            "group_coordinator": self._pykafka_group_coordinator.host,
            "offsets": consumer_offsets,
            "meta": meta,
        }

    def _get_consumer_offsets(self, consumer_id, verbose: bool):
        consumer_offsets = self._unpack_offset_response(
            self._pykafka_group_coordinator.fetch_consumer_group_offsets(consumer_id, preqs=[])
        )

        if verbose:
            for topic in consumer_offsets.keys():
                topic_watermarks = self.topic_controller.get_cluster_topic(topic).watermarks
                for partition_id, consumer_offset in list(consumer_offsets[topic].items()):
                    # TODO somehow include this in the returned dictionary
                    if partition_id not in topic_watermarks:
                        log.warning(
                            f"Found invalid offset! Partition {partition_id} does not exist for topic {topic.decode()}"
                        )
                        del consumer_offsets[topic][partition_id]
                        continue
                    consumer_offsets[topic][partition_id] = {
                        "consumer_offset": consumer_offset,
                        "topic_low_watermark": topic_watermarks[partition_id].low,
                        "topic_high_watermark": topic_watermarks[partition_id].high,
                        "consumer_lag": topic_watermarks[partition_id].high - consumer_offset,
                    }
            return consumer_offsets
        for topic in consumer_offsets.keys():
            topic_watermarks = self.topic_controller.get_cluster_topic(topic).watermarks
            new_consumer_offsets = {
                "consumer_offset": (float("inf"), float("-inf")),
                "topic_low_watermark": (float("inf"), float("-inf")),
                "topic_high_watermark": (float("inf"), float("-inf")),
                "consumer_lag": (float("inf"), float("-inf")),
            }
            for partition_id, consumer_offset in consumer_offsets[topic].items():
                current_offset = consumer_offset
                old_min, old_max = new_consumer_offsets["consumer_offset"]
                new_consumer_offsets["consumer_offset"] = (min(old_min, current_offset), max(old_max, current_offset))

                old_min, old_max = new_consumer_offsets["topic_low_watermark"]
                new_consumer_offsets["topic_low_watermark"] = (
                    min(old_min, topic_watermarks[partition_id].low),
                    max(old_max, topic_watermarks[partition_id].low),
                )

                old_min, old_max = new_consumer_offsets["topic_high_watermark"]
                new_consumer_offsets["topic_high_watermark"] = (
                    min(old_min, topic_watermarks[partition_id].high),
                    max(old_max, topic_watermarks[partition_id].high),
                )

                old_min, old_max = new_consumer_offsets["consumer_lag"]
                new_consumer_offsets["consumer_lag"] = (
                    min(old_min, topic_watermarks[partition_id].high - current_offset),
                    max(old_max, topic_watermarks[partition_id].high - current_offset),
                )

            return new_consumer_offsets

    def _get_member_assignment(self, member_assignment: Dict[str, Any]) -> List[PartitionOffsetFetchRequest]:
        """
        Creates a list of style [PartitionOffsetFetchRequest('topic', partition_id)]
        """
        return [
            PartitionOffsetFetchRequest(topic, partition)
            for member in member_assignment
            for topic, partitions in member["member_assignment"]["partition"].items()
            for partition in partitions
        ]

    def _unpack_offset_response(self, resp: OffsetFetchResponseV1) -> Dict[str, Any]:

        return {
            topic_name: {
                partition_id: partition_data._asdict()["offset"] for partition_id, partition_data in partitions.items()
            }
            for topic_name, partitions in resp.topics.items()
        }

    def _unpack_consumer_group_response(self, resp: DescribeGroupResponse) -> Dict[str, Any]:
        return {
            "group_id": resp.group_id,
            "protocol": resp.protocol,
            "state": resp.state,
            "error_code": resp.error_code,
            "protocol_type": resp.protocol_type,
            "members": [
                {
                    "member_id": member.member_id,
                    "client_id": member.client_id,
                    "client_host": member.client_host,
                    "member_metadata": {
                        # "version": member.member_metadata.version,
                        "subscription": [topic for topic in member.member_metadata.topic_names],
                        # "client": member.member_metadata.user_data,
                    },
                    "member_assignment": {
                        # "version": member.member_assignment.version,
                        "partition": {
                            assignment[0]: assignment[1]
                            for assignment in member.member_assignment.partition_assignment
                        }
                    },
                }
                for _, member in resp.members.items()
            ],
        }



import configparser
import io
import logging
import shutil
from pathlib import Path
from typing import Any, Callable, ClassVar, Dict, List, Optional, Tuple, Type, TypeVar, cast

import yaml

from esque.errors import ConfigTooNew, ConfigTooOld

logger = logging.getLogger(__name__)

CURRENT_VERSION = 1
MIGRATORS: Dict[int, Type["BaseMigrator"]] = {}


def check_config_version(config_path: Path) -> None:
    version = get_config_version(config_path)
    if version < CURRENT_VERSION:
        raise ConfigTooOld(version, CURRENT_VERSION)
    elif version > CURRENT_VERSION:
        raise ConfigTooNew(version, CURRENT_VERSION)


def get_config_version(config_path: Path) -> int:
    if config_path.suffix == ".cfg":
        return 0
    with config_path.open("r") as o:
        return yaml.safe_load(o)["version"]


def migrate(config_path: Path) -> Tuple[Path, Optional[Path]]:
    """
    Migrates the config at `config_path` to the new config version.
    :param config_path: The config to migrate.
    :return: (new path, backup)
    """
    old_version = get_config_version(config_path)
    if old_version == CURRENT_VERSION:
        logger.info("Config already at latest version. Nothing to do.")
        return config_path, None
    logger.info(f"Migrating config from version {old_version} to {CURRENT_VERSION}")

    last_path = config_path
    last_data = None
    migrator = None
    backup = None
    for new_version in range(old_version + 1, CURRENT_VERSION + 1):
        migrator = MIGRATORS[new_version](last_path, last_data)
        if backup is None:
            backup = migrator.backup()
        last_path = migrator.new_config_path
        last_data = migrator.new_data
    migrator.save()
    return last_path, backup


class BaseMigrator:
    VERSION: ClassVar[int] = -1

    def __init__(self, old_config_path: Path, old_data: Any = None):
        self._old_config_path = old_config_path
        self._old_data = old_data
        self.new_data = {}
        if not old_data:
            self._load_old_data()
        self._translate_data()

    def _load_old_data(self):
        self.deserialize(self._old_config_path.read_text())

    def _translate_data(self):
        self.new_data["version"] = self.VERSION

    @property
    def new_config_path(self) -> Path:
        return self._old_config_path

    def serialize(self) -> str:
        raise NotImplementedError()

    def deserialize(self, text: str):
        raise NotImplementedError()

    def backup(self) -> Path:
        backup = self._old_config_path.with_name(self._old_config_path.name + ".bak")
        shutil.copy(str(self._old_config_path), str(backup))
        return backup

    def save(self) -> Path:
        self.new_config_path.write_text(self.serialize())
        return self.new_config_path


class V1Migrator(BaseMigrator):
    VERSION = 1

    @property
    def new_config_path(self) -> Path:
        return self._old_config_path.parent / "esque_config.yaml"

    def deserialize(self, text: str):
        self._old_data = configparser.ConfigParser()
        self._old_data.read_string(text)

    def serialize(self) -> str:
        buffer = io.StringIO()
        yaml.dump(self.new_data, buffer, default_flow_style=False, sort_keys=False, Dumper=yaml.SafeDumper)
        return buffer.getvalue()

    def _translate_data(self):
        super()._translate_data()
        self.new_data["current_context"] = self._old_data["Context"]["current"]
        self._translate_contexts()

    def _translate_contexts(self):
        all_contexts = filter(V1Migrator.is_context, self._old_data.values())
        translated_contexts = dict(map(V1Migrator.translate_context, all_contexts))
        self.new_data["contexts"] = translated_contexts

    @staticmethod
    def is_context(section: configparser.SectionProxy) -> bool:
        return section.name.startswith("Context.")

    @staticmethod
    def translate_context(section: configparser.SectionProxy) -> Tuple[str, Dict]:
        data = {"bootstrap_servers": V1Migrator.translate_bootstrap_servers(section)}
        unchanged_fields = ["security_protocol", "schema_registry"]
        for field in unchanged_fields:
            assign_if_present(field, cast(Dict[str, Any], section), data)

        data.update(V1Migrator.get_ssl_settings(section))
        data.update(V1Migrator.get_sasl_settings(section))
        data.update(V1Migrator.translate_defaults(section))

        name = V1Migrator.extract_name_from_section(section)
        return name, data

    @staticmethod
    def translate_defaults(section: configparser.SectionProxy) -> Dict:
        data = {}
        if "default_replication_factor" in section:
            data["replication_factor"] = int(section["default_replication_factor"])
        if "default_partitions" in section:
            data["num_partitions"] = int(section["default_partitions"])

        return {"default_values": data} if data != {} else data

    @staticmethod
    def translate_bootstrap_servers(section: configparser.SectionProxy) -> List[str]:
        server_tpl = "{}"
        if section.get("bootstrap_domain"):
            server_tpl += "." + section["bootstrap_domain"]
        if section.get("bootstrap_port"):
            server_tpl += ":" + section["bootstrap_port"]
        return [server_tpl.format(server.strip()) for server in section["bootstrap_hosts"].split(",")]

    @staticmethod
    def get_ssl_settings(section: configparser.SectionProxy) -> Dict:
        has_ssl_entries = any(entry.startswith("ssl_") for entry in section.keys())
        if not has_ssl_entries:
            return {}

        def translate(ssl_param_name: str) -> str:
            return ssl_param_name[4:]

        ssl_params = ["ssl_cafile", "ssl_certfile", "ssl_keyfile", "ssl_password"]
        data = {}
        for param in ssl_params:
            assign_if_present(param, cast(Dict[str, Any], section), data, translate)
        return {"ssl_params": data}

    @staticmethod
    def get_sasl_settings(section: configparser.SectionProxy) -> Dict:
        has_sasl_entries = any(entry.startswith("sasl_") for entry in section.keys())
        if not has_sasl_entries:
            return {}

        def translate(sasl_param_name: str) -> str:
            return sasl_param_name[5:]

        sasl_params = ["sasl_mechanism", "sasl_user", "sasl_password"]

        data = {translate(param): section[param] for param in sasl_params}
        return {"sasl_params": data}

    @staticmethod
    def extract_name_from_section(section: configparser.SectionProxy) -> str:
        return section.name.split(".", 1)[1]


MIGRATORS[1] = V1Migrator

K = TypeVar("K")


def identity(x: K) -> K:
    return x


def assign_if_present(
    key: K, sourcedict: Dict[K, Any], targetdict: Dict[K, Any], translate: Callable[[K], K] = identity
) -> bool:
    """
    Assigns `sourcedict[key]` to `targetdict[translate(key)]` if and only if sourcedict contains `key`.
    :param key: key to check for
    :param sourcedict: source to check in
    :param targetdict: target to assign to
    :param translate: defaults to identity i.e no translation. Can be used to change key in target dict.
    :return: bool, `key in sourcedict`
    """
    if key in sourcedict:
        targetdict[translate(key)] = sourcedict[key]
        return True
    return False



import logging
import random
import string
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional

import click
import yaml
from confluent_kafka.admin import ConfigResource
from pykafka import SslConfig

import esque.validation
from esque.cli import environment
from esque.config.migration import check_config_version
from esque.errors import (
    ConfigException,
    ConfigNotExistsException,
    ContextNotDefinedException,
    ExceptionWithMessage,
    MissingSaslParameter,
    UnsupportedSaslMechanism,
)
from esque.helpers import SingletonMeta

if TYPE_CHECKING:
    try:
        from pykafka.sasl_authenticators import BaseAuthenticator
    except ImportError:
        raise ImportError(
            "Please install our pykafka fork:\n"
            "pip install -U git+https://github.com/real-digital/pykafka.git@feature/sasl-scram-support"
        )

RANDOM = "".join(random.choices(string.ascii_lowercase, k=8))
PING_TOPIC = f"ping-{RANDOM}"
PING_GROUP_ID = f"ping-{RANDOM}"
SLEEP_INTERVAL = 2
SUPPORTED_SASL_MECHANISMS = ("PLAIN", "SCRAM-SHA-256", "SCRAM-SHA-512")
log = logging.getLogger(__name__)


def config_dir() -> Path:
    return _config_dir()


def config_path() -> Path:
    return _config_path()


# create functions we can mock during tests
def _config_path() -> Path:
    if environment.ESQUE_CONF_PATH:
        return Path(environment.ESQUE_CONF_PATH)
    legacy_path = config_dir() / "esque.cfg"
    current_path = config_dir() / "esque_config.yaml"
    if legacy_path.exists() and not current_path.exists():
        return legacy_path
    return current_path


def _config_dir() -> Path:
    return Path(click.get_app_dir("esque", force_posix=True))


def sample_config_path() -> Path:
    return Path(__file__).parent / "sample_config.yaml"


class Config(metaclass=SingletonMeta):
    def __init__(self, *, disable_validation=False):
        if not config_path().exists():
            raise ConfigNotExistsException()
        check_config_version(config_path())
        self._cfg = yaml.safe_load(config_path().read_text())
        if not disable_validation:
            esque.validation.validate_esque_config(self._cfg)
        self._current_dict: Optional[Dict[str, str]] = None

    @property
    def available_contexts(self) -> List[str]:
        return sorted(self._cfg["contexts"])

    @property
    def default_values(self) -> Dict[str, Any]:
        return self.current_context_dict.get("default_values", {})

    @property
    def current_context(self) -> str:
        return self._cfg["current_context"]

    @property
    def current_context_dict(self) -> Dict[str, Any]:
        return self._cfg["contexts"][self.current_context]

    @property
    def schema_registry(self) -> str:
        config_dict = self.current_context_dict
        return config_dict["schema_registry"]

    @property
    def bootstrap_servers(self) -> List[str]:
        return self.current_context_dict["bootstrap_servers"]

    @property
    def default_num_partitions(self) -> int:
        if "num_partitions" in self.default_values:
            return self.default_values["num_partitions"]

        from esque.cli.output import bold, blue_bold

        try:
            log.warning("Fetching default number of partitions from broker.")
            log.warning(f"Run `esque config edit` and add `{bold('num_partitions')}` to your defaults to avoid this.")
            default = int(self._get_broker_setting("num.partitions"))
            log.warning(f"Cluster default is {blue_bold(str(default))}.")
        except Exception as e:
            default = 1
            log.warning(f"Fetching default from broker failed. Falling back to {default}")
            log.info(type(e).__name__, exc_info=True)
        return default

    @staticmethod
    def _get_broker_setting(self, setting: str) -> str:
        from esque.cluster import Cluster

        cluster = Cluster()
        brokers = cluster.brokers
        config = cluster.retrieve_config(ConfigResource.Type.BROKER, brokers[0]["id"])
        return config[setting]

    @property
    def default_replication_factor(self) -> int:
        if "replication_factor" in self.default_values:
            return self.default_values["replication_factor"]

        from esque.cli.output import bold, blue_bold

        try:
            log.warning("Fetching default replication factor from broker.")
            log.warning(
                f"Run `esque config edit` and add `{bold('replication_factor')}` to your defaults to avoid this."
            )
            default = int(self._get_broker_setting("default.replication.factor"))
            log.warning(f"Cluster default is {blue_bold(str(default))}.")
        except Exception as e:
            default = min(len(self.bootstrap_servers), 3)
            log.warning(f"Fetching default from broker failed. Falling back to {default}")
            log.info(type(e).__name__, exc_info=True)
        return default

    @property
    def sasl_mechanism(self) -> str:
        if "mechanism" not in self.sasl_params:
            raise MissingSaslParameter(f"No sasl mechanism configured, valid values are {SUPPORTED_SASL_MECHANISMS}")
        return self.sasl_params["mechanism"].upper()

    @property
    def sasl_enabled(self) -> bool:
        return "SASL" in self.security_protocol

    @property
    def sasl_params(self) -> Dict[str, str]:
        return self.current_context_dict.get("sasl_params", {})

    @property
    def ssl_params(self) -> Dict[str, str]:
        return self.current_context_dict.get("ssl_params", {})

    @property
    def ssl_enabled(self) -> bool:
        return "SSL" in self.security_protocol

    @property
    def security_protocol(self) -> str:
        protocol = self.current_context_dict.get("security_protocol", "PLAINTEXT").upper()
        if "SASL" in protocol and "sasl_params" not in self.current_context_dict:
            msg = (
                f"Security protocol {protocol} contains 'SASL' indicating that you want to connect to "
                "a SASL enabled endpoint but you didn't specify sasl parameters.\n"
                f"Run `esque config edit` and add `sasl_params` to section 'contexts.{self.current_context}'.\n"
                f"Valid mechanisms are {SUPPORTED_SASL_MECHANISMS}."
            )
            raise ConfigException(msg)
        return protocol

    def context_switch(self, context: str):
        if context not in self.available_contexts:
            raise ContextNotDefinedException(f"{context} not defined in {config_path()}")
        self._cfg["current_context"] = context

    def save(self):
        with config_path().open("w") as f:
            yaml.dump(self._cfg, f, default_flow_style=False, sort_keys=False, Dumper=yaml.SafeDumper)

    def create_pykafka_config(self) -> Dict[str, Any]:
        config = {"hosts": ",".join(self.bootstrap_servers), "exclude_internal_topics": False}
        if self.sasl_enabled:
            config["sasl_authenticator"] = self._get_pykafka_authenticator()
        if self.ssl_enabled:
            config["ssl_config"] = self._get_pykafka_ssl_conf()
        log.debug(f"Created pykafka config: {config}")
        return config

    def create_confluent_config(self, *, debug: bool = False, include_schema_registry: bool = False) -> Dict[str, str]:
        config = {"bootstrap.servers": ",".join(self.bootstrap_servers), "security.protocol": self.security_protocol}
        if debug:
            config.update({"debug": "all", "log_level": "2"})

        if include_schema_registry:
            config["schema.registry.url"] = self.schema_registry

        if self.sasl_enabled:
            config.update(self._get_confluent_sasl_config())
        if self.ssl_enabled:
            config.update(self._get_confluent_ssl_config())
        log.debug(f"Created confluent config: {config}")
        return config

    def _get_confluent_sasl_config(self) -> Dict[str, str]:
        if self.sasl_mechanism in ("PLAIN", "SCRAM-SHA-256", "SCRAM-SHA-512"):
            try:
                return {
                    "sasl.mechanisms": self.sasl_mechanism,
                    "sasl.username": self.sasl_params["user"],
                    "sasl.password": self.sasl_params["password"],
                }
            except KeyError as e:
                msg = f"SASL mechanism {self.sasl_mechanism} requires parameter {e.args[0]}.\n"
                msg += f"Please run `esque config edit` and add it to section [{self.current_context}]."
                raise MissingSaslParameter(msg)
        else:
            raise UnsupportedSaslMechanism(
                f"SASL mechanism {self.sasl_mechanism} is currently not supported by esque. "
                f"Supported meachnisms are {SUPPORTED_SASL_MECHANISMS}."
            )

    def _get_confluent_ssl_config(self) -> Dict[str, str]:
        if not self.ssl_enabled:
            return {}
        rdk_conf = {}
        if self.ssl_params.get("cafile"):
            rdk_conf["ssl.ca.location"] = self.ssl_params["cafile"]
        if self.ssl_params.get("certfile"):
            rdk_conf["ssl.certificate.location"] = self.ssl_params["certfile"]
        if self.ssl_params.get("keyfile"):
            rdk_conf["ssl.key.location"] = self.ssl_params["keyfile"]
        if self.ssl_params.get("password"):
            rdk_conf["ssl.key.password"] = self.ssl_params["password"]
        return rdk_conf

    def _get_pykafka_ssl_conf(self) -> Optional[SslConfig]:
        if not self.ssl_enabled:
            return None
        ssl_params = {"cafile": self.ssl_params.get("cafile", None)}
        if self.ssl_params.get("certfile"):
            ssl_params["certfile"] = self.ssl_params["certfile"]
        if self.ssl_params.get("keyfile"):
            ssl_params["keyfile"] = self.ssl_params["keyfile"]
        if self.ssl_params.get("password"):
            ssl_params["password"] = self.ssl_params["password"]
        return SslConfig(**ssl_params)

    def _get_pykafka_authenticator(self) -> "BaseAuthenticator":
        try:
            from pykafka.sasl_authenticators import PlainAuthenticator, ScramAuthenticator
        except ImportError:
            raise ExceptionWithMessage(
                "In order to support SASL you'll need to install our fork of pykafka.\n"
                "Please run:\n"
                "    pip install -U git+https://github.com/real-digital/pykafka.git@feature/sasl-scram-support"
            )

        try:
            if self.sasl_mechanism == "PLAIN":
                return PlainAuthenticator(
                    user=self.sasl_params["user"],
                    password=self.sasl_params["password"],
                    security_protocol=self.security_protocol,
                )
            elif self.sasl_mechanism in ("SCRAM-SHA-256", "SCRAM-SHA-512"):
                return ScramAuthenticator(
                    self.sasl_mechanism,
                    user=self.sasl_params["user"],
                    password=self.sasl_params["password"],
                    security_protocol=self.security_protocol,
                )
            else:
                raise UnsupportedSaslMechanism(
                    f"SASL mechanism {self.sasl_mechanism} is currently not supported by esque. "
                    f"Supported meachnisms are {SUPPORTED_SASL_MECHANISMS}."
                )
        except KeyError as e:
            msg = f"SASL mechanism {self.sasl_mechanism} requires parameter {e.args[0]}.\n"
            msg += f"Please run `esque config edit` and add it to section [{self.current_context}]."
            raise MissingSaslParameter(msg)



import logging
import re
from typing import Dict, List, Optional, Tuple

import pendulum
import pykafka
from confluent_kafka.cimpl import TopicPartition

from esque.cli.output import red_bold
from esque.clients.consumer import ConsumerFactory
from esque.cluster import Cluster
from esque.controller.topic_controller import TopicController
from esque.resources.consumergroup import ConsumerGroup


class ConsumerGroupOffsetPlan:
    def __init__(
        self,
        topic_name: str,
        current_offset: int,
        proposed_offset: int,
        high_watermark: int,
        low_watermark: int,
        partition_id: int,
    ) -> None:
        self.topic_name = topic_name
        self.current_offset = current_offset
        self.proposed_offset = proposed_offset
        self.high_watermark = high_watermark
        self.low_watermark = low_watermark
        self.partition_id = partition_id

    @property
    def offset_equal(self) -> bool:
        return self.current_offset == self.proposed_offset


class ConsumerGroupController:
    def __init__(self, cluster: Cluster):
        self.cluster = cluster
        self._logger = logging.getLogger(__name__)

    def get_consumergroup(self, consumer_id) -> ConsumerGroup:
        return ConsumerGroup(consumer_id, self.cluster)

    def list_consumer_groups(self) -> List[str]:
        brokers: Dict[int, pykafka.broker.Broker] = self.cluster.pykafka_client.cluster.brokers
        return list(
            set(group.decode("UTF-8") for _, broker in brokers.items() for group in broker.list_groups().groups)
        )

    def edit_consumer_group_offsets(self, consumer_id: str, offset_plan: List[ConsumerGroupOffsetPlan]):
        """
        Commit consumergroup offsets to specific values
        :param consumer_id: ID of the consumer group
        :param offset_plan: List of ConsumerGroupOffsetPlan objects denoting the offsets for each partition in different topics
        :return:
        """
        consumer = ConsumerFactory().create_consumer(
            group_id=consumer_id,
            topic_name=None,
            output_directory=None,
            last=False,
            avro=False,
            initialize_default_output_directory=False,
            match=None,
            enable_auto_commit=False,
        )

        offsets = [
            TopicPartition(
                topic=plan_element.topic_name, partition=plan_element.partition_id, offset=plan_element.proposed_offset
            )
            for plan_element in offset_plan
            if not plan_element.offset_equal
        ]
        consumer.commit(offsets=offsets)

    def create_consumer_group_offset_change_plan(
        self,
        consumer_id: str,
        topic_name: str,
        offset_to_value: Optional[int],
        offset_by_delta: Optional[int],
        offset_to_timestamp: Optional[str],
        offset_from_group: Optional[str],
    ) -> List[ConsumerGroupOffsetPlan]:

        consumer_group_state, offset_plans = self._read_current_consumergroup_offsets(
            consumer_id=consumer_id, topic_name_expression=topic_name
        )
        if consumer_group_state == "Dead":
            self._logger.error("The consumer group {} does not exist.".format(consumer_id))
            return None
        elif consumer_group_state == "Empty":
            if offset_to_value is not None:
                for plan_element in offset_plans.values():
                    (allowed_offset, error, message) = self._select_new_offset_for_consumer(
                        offset_to_value, plan_element
                    )
                    plan_element.proposed_offset = allowed_offset
                    if error:
                        self._logger.error(message)
            elif offset_by_delta is not None:
                for plan_element in offset_plans.values():
                    requested_offset = plan_element.current_offset + offset_by_delta
                    (allowed_offset, error, message) = self._select_new_offset_for_consumer(
                        requested_offset, plan_element
                    )
                    plan_element.proposed_offset = allowed_offset
                    if error:
                        self._logger.error(message)
            elif offset_to_timestamp is not None:
                timestamp_limit = pendulum.parse(offset_to_timestamp)
                proposed_offset_dict = TopicController(self.cluster, None).get_offsets_closest_to_timestamp(
                    group_id=consumer_id, topic_name=topic_name, timestamp_limit=timestamp_limit
                )
                for plan_element in offset_plans.values():
                    plan_element.proposed_offset = proposed_offset_dict.get(plan_element.partition_id, 0)
            elif offset_from_group is not None:
                _, mirror_consumer_group = self._read_current_consumergroup_offsets(
                    consumer_id=offset_from_group, topic_name_expression=topic_name
                )
                for key, value in mirror_consumer_group.items():
                    if key in offset_plans.keys():
                        offset_plans[key].proposed_offset = value.current_offset
                    else:
                        value.current_offset = 0
                        offset_plans[key] = value
            return list(offset_plans.values())
        else:
            self._logger.error(
                "Consumergroup {} is not empty. Use the {} option if you want to override this safety mechanism.".format(
                    consumer_id, red_bold("--force")
                )
            )
            return list(offset_plans.values())

    @staticmethod
    def _select_new_offset_for_consumer(
        requested_offset: int, offset_plan: ConsumerGroupOffsetPlan
    ) -> Tuple[int, bool, str]:
        if requested_offset < offset_plan.low_watermark:
            final_value = offset_plan.low_watermark
            error = True
            message = "The requested offset ({}) is outside of the allowable range [{},{}] for partition {} in topic {}. Setting to low watermark.".format(
                requested_offset,
                red_bold(str(offset_plan.low_watermark)),
                offset_plan.high_watermark,
                offset_plan.partition_id,
                offset_plan.topic_name,
            )
        elif requested_offset > offset_plan.high_watermark:
            final_value = offset_plan.high_watermark
            error = True
            message = "The requested offset ({}) is outside of the allowable range [{},{}] for partition {} in topic {}. Setting to high watermark.".format(
                requested_offset,
                offset_plan.low_watermark,
                red_bold(str(offset_plan.high_watermark)),
                offset_plan.partition_id,
                offset_plan.topic_name,
            )
        else:
            final_value = requested_offset
            error = False
            message = ""
        return final_value, error, message

    def _read_current_consumergroup_offsets(
        self, consumer_id: str, topic_name_expression: str
    ) -> Tuple[str, Dict[str, ConsumerGroupOffsetPlan]]:
        offset_plans = {}
        topic_name_pattern = re.compile(topic_name_expression, re.IGNORECASE)
        try:
            consumer_group = self.get_consumergroup(consumer_id)
            consumer_group_desc = consumer_group.describe(verbose=True)
            consumer_group_state = consumer_group_desc["meta"]["state"].decode("UTF-8")
            for subscribed_topic_name in consumer_group_desc["offsets"]:
                decoded_topic_name = subscribed_topic_name.decode("UTF-8")
                if topic_name_pattern.match(decoded_topic_name):
                    for partition_id, partition_info in consumer_group_desc["offsets"][subscribed_topic_name].items():
                        consumer_offset_plan = ConsumerGroupOffsetPlan(
                            topic_name=decoded_topic_name,
                            current_offset=partition_info["consumer_offset"],
                            proposed_offset=partition_info["consumer_offset"],
                            high_watermark=partition_info["topic_high_watermark"],
                            low_watermark=partition_info["topic_low_watermark"],
                            partition_id=partition_id,
                        )
                        offset_plans[f"{decoded_topic_name}::{partition_id}"] = consumer_offset_plan
        except AttributeError:
            return "Dead", offset_plans
        if len(offset_plans) == 0:
            self._logger.warning("No offsets have ever been committed by consumergroup {}.".format(consumer_id))
        return consumer_group_state, offset_plans



from pathlib import Path
from typing import Dict

import yamale

from esque.errors import ValidationException, YamaleValidationException
from esque.validation import yamale_validators

SCHEMA_DIR = Path(__file__).resolve().parent / "schemas"


def validate_topic_config(topic_config: Dict) -> None:
    validate(topic_config, SCHEMA_DIR / "topic.yaml")


def validate_editable_topic_config(editable_topic_config: Dict) -> None:
    validate(editable_topic_config, SCHEMA_DIR / "editable_topic.yaml")


def validate_esque_config(esque_config: Dict) -> None:
    validate(esque_config, SCHEMA_DIR / "esque_config.yaml")
    ctx = esque_config["current_context"]
    all_ctx = sorted(esque_config["contexts"])
    if ctx not in all_ctx:
        raise ValidationException(f"Context {ctx} does not exist. Available contexts {all_ctx}")


def validate(data: Dict, schema: Path) -> None:
    schema = yamale.make_schema(schema, validators=yamale_validators.all_validators())
    try:
        yamale.validate(schema, [(data, "<dict literal>")], strict=True)
    except ValueError as validation_error:
        raise YamaleValidationException(validation_error)



import pathlib
import shutil
from typing import Callable, Dict, Optional, Tuple, Union

import click
import yaml
from yaml.scanner import ScannerError

from esque.errors import EditCanceled, NoConfirmationPossibleException, ValidationException


# private function, which we can mock
def _isatty(stream) -> bool:
    return stream.isatty()


def isatty(stream) -> bool:
    return _isatty(stream)


def ensure_approval(question: str, *, no_verify: bool = False, default_answer=False) -> bool:
    if no_verify:
        return True

    if not isatty(click.get_text_stream("stdin")):
        raise NoConfirmationPossibleException()

    return click.confirm(question, default=default_answer)


def attrgetter(*attrs):
    if len(attrs) == 1:

        def getter(obj):
            return getattr(obj, attrs[0])

    else:

        def getter(obj):
            return tuple(getattr(obj, attr) for attr in attrs)

    return getter


class HandleFileOnFinished:
    def __init__(self, dir_: pathlib.Path, keep_file: bool):
        self.keep_file = keep_file
        self._dir = dir_
        self._dir.mkdir(parents=True)

    def __enter__(self) -> pathlib.Path:
        return self._dir

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.keep_file and self._dir.exists():
            shutil.rmtree(self._dir)


def edit_yaml(yaml_str: str, validator: Optional[Callable[[Dict], None]] = None) -> Tuple[str, Dict]:
    while True:
        yaml_str: Optional[str] = click.edit(yaml_str, extension=".yaml")

        # edit process can be aborted, ex. in vim via :q!
        if yaml_str is None:
            raise EditCanceled()
        try:
            # TODO: check for duplicate keys, might have to change yaml parser for that
            config_data = yaml.safe_load(yaml_str)
            if validator:
                validator(config_data)
        except (ScannerError, ValidationException) as e:
            _handle_edit_exception(e)
        else:
            break
    return yaml_str, config_data


def _handle_edit_exception(e: Union[ScannerError, ValidationException]) -> None:
    if isinstance(e, ScannerError):
        click.echo("Error parsing yaml:")
    else:
        click.echo("Error validating yaml:")
    click.echo(str(e))
    if not ensure_approval("Continue Editing?", default_answer=True):
        raise EditCanceled()



import os

# Make sure to only ever import the whole module instead of single variables
# That way it is easier to mock the variables during testing if necessary.
# BAD: from esque.cli.environment import ESQUE_CONF_PATH
# GOOD: from esque.cli import environment

ESQUE_CONF_PATH = os.environ.get("ESQUE_CONF_PATH")
ESQUE_VERBOSE = os.environ.get("ESQUE_VERBOSE")



import json
import logging
import pathlib
import sys
from abc import ABC, abstractmethod
from json import JSONDecodeError
from typing import Optional

import confluent_kafka
import pendulum
from confluent_kafka.avro import AvroProducer

from esque.config import Config
from esque.errors import raise_for_kafka_error
from esque.helpers import delta_t, log_error
from esque.messages.avromessage import AvroFileReader
from esque.messages.message import (
    FileReader,
    KafkaMessage,
    PlainTextFileReader,
    deserialize_message,
    get_partitions_in_path,
)
from esque.ruleparser.ruleengine import RuleTree


class AbstractProducer(ABC):
    def __init__(self, topic_name: str, match: str = None):
        self.queue_length = 100000
        self.internal_queue_length_limit = self.queue_length / 0.5
        self._config = Config.get_instance().create_confluent_config()
        self._setup_config()
        self.logger = logging.getLogger(__name__)
        self._topic_name = topic_name
        self._match = match
        self._producer = None
        if self._match is not None:
            self._rule_tree = RuleTree(match)
        else:
            self._rule_tree = None
        self.create_internal_producer()

    @abstractmethod
    def produce(self) -> int:
        raise NotImplementedError()

    def _setup_config(self):
        self._config.update(
            {
                "on_delivery": raise_for_kafka_error,
                "error_cb": log_error,
                "queue.buffering.max.messages": str(self.queue_length),
            }
        )

    def flush_all(self, message_prefix: str = None):
        while True:
            left_messages = self._producer.flush(1)
            if left_messages == 0:
                break
            self.logger.info((message_prefix or "") + f"Still {left_messages} messages left, flushing...")

    def create_internal_producer(self):
        self._producer = confluent_kafka.Producer(self._config)

    def produce_message(self, topic_name: str, message: KafkaMessage):
        if self._rule_tree is None or self._rule_tree.evaluate(message):
            self._producer.produce(
                topic=topic_name,
                key=message.key,
                value=message.value,
                partition=message.partition,
                headers=message.headers,
            )


class PingProducer(AbstractProducer):
    def produce(self) -> int:
        start = pendulum.now()
        self.produce_message(
            topic_name=self._topic_name,
            message=KafkaMessage(key=str(0), value=str(pendulum.now().timestamp()), partition=0),
        )
        self.flush_all(message_prefix=f"{delta_t(start)} | ")
        return 1


class StdInProducer(AbstractProducer):
    def __init__(self, topic_name: str, match: str = None, ignore_errors: bool = False):
        super().__init__(topic_name=topic_name, match=match)
        self._ignore_errors = ignore_errors

    def produce(self) -> int:
        total_number_of_produced_messages = 0
        # accept lines from stdin until EOF
        for single_message_line in sys.stdin:
            message = None
            try:
                message = deserialize_message(single_message_line.replace("\n", ""))
            except (JSONDecodeError, KeyError):
                if self._ignore_errors:
                    message = KafkaMessage(key=None, value=single_message_line, partition=0)
            if message:
                self.produce_message(topic_name=self._topic_name, message=message)
                total_number_of_produced_messages += 1
        return total_number_of_produced_messages


class FileProducer(AbstractProducer):
    def __init__(self, topic_name: str, input_directory: pathlib.Path, match: str = None):
        super().__init__(topic_name=topic_name, match=match)
        self.input_directory = input_directory

    def produce(self) -> int:
        partitions = get_partitions_in_path(self.input_directory)
        counter = 0
        for partition in partitions:
            with self.get_file_reader(pathlib.Path(self.input_directory), partition) as file_reader:
                for message in file_reader.read_message_from_file():
                    self.produce_message(self._topic_name, message)
                    left_messages = self._producer.flush(0)
                    if left_messages > self.internal_queue_length_limit:
                        self.flush_all()
                    counter += 1
                self.flush_all()
        return counter

    def get_file_reader(self, directory: pathlib.Path, partition: Optional[str]) -> FileReader:
        return PlainTextFileReader(directory, partition)


class AvroFileProducer(FileProducer):
    def create_internal_producer(self):
        self._producer = AvroProducer(self._config)

    def get_file_reader(self, directory: pathlib.Path, partition: Optional[str]) -> FileReader:
        return AvroFileReader(directory, partition)

    def _setup_config(self):
        super()._setup_config()
        self._config.update({"schema.registry.url": Config.get_instance().schema_registry})

    def produce_message(self, topic_name: str, message: KafkaMessage):
        if self._rule_tree is None or self._rule_tree.evaluate(message):
            self._producer.produce(
                topic=topic_name,
                key=None if message.key is None else json.loads(message.key),
                value=None if message.value is None else json.loads(message.value),
                partition=message.partition,
                key_schema=message.key_schema,
                value_schema=message.value_schema,
            )


class ProducerFactory:
    def create_producer(
        self,
        topic_name: str,
        input_directory: pathlib.Path,
        avro: bool,
        match: str = None,
        ignore_stdin_errors: bool = False,
    ):
        if input_directory is None:
            producer = StdInProducer(topic_name=topic_name, match=match, ignore_errors=ignore_stdin_errors)
        elif avro:
            producer = AvroFileProducer(topic_name=topic_name, input_directory=input_directory, match=match)
        else:
            producer = FileProducer(topic_name=topic_name, input_directory=input_directory, match=match)
        return producer
